#!/usr/bin/env python

import sys
import types
import importlib

import pywbem


def main():

    try:
        version = pywbem.__version__
    except AttributeError:
        try:
            version = pywbem._version
        except AttributeError:
            version = '?'

    with open('try/submodule_symbols_0_12_0.txt') as fp:
        lines = fp.readlines()

    print("Submodule symbols from 0.12.0 that did not exist in version %s:" %
          version)

    new_submods = set()
    for line in lines:
        submod_name, kind, sym_name = line.split()

        try:
            submod = getattr(pywbem, submod_name)
        except AttributeError:
            # submodule not automatically imported when main module imported
            # (e.g. mof_compiler before pywbem 0.10)
            try:
                importlib.import_module('pywbem.%s' % submod_name)
            except ImportError:
                new_submods.add(submod_name)
                continue
            submod = getattr(pywbem, submod_name)

        if not hasattr(submod, sym_name):
            print("  %s.%s (%s)" % (submod_name, sym_name, kind))

    if new_submods:
        print("Submodules that did not exist yet in version %s:" %
              version)
    for name in new_submods:
        print("  %s" % name)

    return 0


if __name__ == '__main__':
    sys.exit(main())



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
    Test wbemcli script.  This test generates a cmdline that calls
    wbemcli with a specific set of options and tests the returns.
    Because wbemcli always goes to interactive mode, the test call
    includes a wbemcli script that forces wbemcli to quit.

    It dynamically generates the set of tests from the TEST_MAP table
    where each test is a call to execute wbemcli with a particular set of
    arguments and options. It then tests the stdout, stderr, and exitcode
    against the TEST_MAP.

    This test does not test the individual operations that can be executed,
    just the initialization components for wbemcli
"""

from __future__ import print_function, absolute_import
import os
import pytest

from .utils.cmd_line_test_utils import wbemcli_test

TEST_DIR = os.path.dirname(__file__)
TEST_MOCK_MOF = os.path.join(TEST_DIR, 'simple_mock_model.mof')

VERBOSE = False  # Verbose mode for the testcases


def create_abs_path(filename):
    """
    Create an absolute path name for filename in the same directory as this
    python code.
    """
    script_dir = os.path.dirname(__file__)
    return os.path.join(script_dir, filename)


WBEMCLI_HELP = """usage: wbemcli [options] server

Provide an interactive shell for issuing operations against a WBEM server.

wbemcli executes the WBEMConnection as part of initialization so the user can
input requests as soon as the interactive shell is started.

Use h() in thenteractive shell for help for wbemcli methods and variables.

Positional arguments:
  server                Host name or url of the WBEM server in this format:
                            [{scheme}://]{host}[:{port}]
                        - scheme: Defines the protocol to use;
                            - "https" for HTTPs protocol
                            - "http" for HTTP protocol.
                          Default: "https".
                        - host: Defines host name as follows:
                             - short or fully qualified DNS hostname,
                             - literal IPV4 address(dotted)
                             - literal IPV6 address (RFC 3986) with zone
                               identifier extensions(RFC 6874)
                               supporting "-" or %25 for the delimiter.
                        - port: Defines the WBEM server port to be used
                          Defaults:
                             - HTTP  - 5988
                             - HTTPS - 5989

Server related options:
  Specify the WBEM server namespace and timeout

  -n namespace, --namespace namespace
                        Default namespace in the WBEM server for operation
                        requests when namespace option not supplied with
                        operation request.
                        Default: root/cimv2
  -t timeout, --timeout timeout
                        Timeout of the completion of WBEM Server operation
                        in seconds(integer between 0 and 300).
                        Default: No timeout

Connection security related options:
  Specify user name and password or certificates and keys

  -u user, --user user  User name for authenticating with the WBEM server.
                        Default: No user name.
  -p password, --password password
                        Password for authenticating with the WBEM server.
                        Default: Will be prompted for, if user name
                        specified.
  -nvc, --no-verify-cert
                        Client will not verify certificate returned by the
                        WBEM server (see cacerts). This bypasses the client-
                        side verification of the server identity, but allows
                        encrypted communication with a server for which the
                        client does not have certificates.
  --cacerts cacerts     File or directory containing certificates that will be
                        matched against a certificate received from the WBEM
                        server. Set the --no-verify-cert option to bypass
                        client verification of the WBEM server certificate.
                        Default: Searches for matching certificates in the
                        following system directories:
                        /etc/pki/ca-trust/extracted/openssl/ca-bundle.trust.crt
                        /etc/ssl/certs
                        /etc/ssl/certificates
  --certfile certfile   Client certificate file for authenticating with the
                        WBEM server. If option specified the client attempts
                        to execute mutual authentication.
                        Default: Simple authentication.
  --keyfile keyfile     Client private key file for authenticating with the
                        WBEM server. Not required if private key is part of the
                        certfile option. Not allowed if no certfile option.
                        Default: No client key file. Client private key should
                        then be part  of the certfile

General options:
  -s [scripts [scripts ...]], --scripts [scripts [scripts ...]]
                        Execute the python code defined by the script before the
                        user gets control. This argument may be repeated to load
                        multiple scripts or multiple scripts may be listed for a
                        single use of the option. Scripts are executed after the
                        WBEMConnection call
  -v, --verbose         Print more messages while processing
  -V, --version         Display pywbem version and exit.
  --statistics          Enable gathering of statistics on operations.
  --mock-server [file name [file name ...]]
                        Activate pywbem_mock in place of a live WBEMConnection and
                        compile/build the files defined (".mof" suffix or "py" suffix.
                        MOF files are compiled and python files are executed assuming
                        that they include mock_pywbem methods that add objects to the
                        repository.
  -l log_spec[,logspec], --log log_spec[,logspec]
                        Log_spec defines characteristics of the various named
                        loggers. It is the form:
                         COMP=[DEST[:DETAIL]] where:
                           COMP:   Logger component name:[api|http|all].
                                   (Default=all)
                           DEST:   Destination for component:[file|stderr].
                                   (Default=file)
                           DETAIL: Detail Level to log: [all|paths|summary] or
                                   an integer that defines the maximum length of
                                   of each log record.
                                   (Default=all)
  -h, --help            Show this help message and exit

Examples:
  wbemcli https://localhost:15345 -n vendor -u sheldon -p penny
          - (https localhost, port=15345, namespace=vendor user=sheldon
         password=penny)

  wbemcli http://[2001:db8::1234-eth0] -(http port 5988 ipv6, zone id eth0)
"""  # noqa=E501

# pylint: disable=line-too-long


TEST_SCRIPT = """
#
#  This script tests a number of the WBEM operation methods as defined
#  for wbemcli. It executes the commands and then asserts for the expected
#  results.  This script tests the enumerate and get for class, instance, and
#  qualifier operations.
#
from wbemcli import CONN, gc, ecn, ei, iei, oei, piwp, ein, ieip, oeip, pip, \
    a, iai, oai, an, iaip, oaip, iri, ori, rn, irip, orip, ec, gq, eq

print('CONN %s' % CONN)

qual_decls = eq()
# print('QUALIFIER_DECLARATIONS %s' % qual_decls)
assert 'Association' in [qd.name for qd in qual_decls]

qualifier = gq('Association')
# print('QUALIFIER %s' % qualifier)
assert qualifier.name == 'Association'

class_names = ecn()
# print('CLASSNAMES %s' % class_names)
assert class_names == ['CIM_Foo']

clns = ecn(di=True)
assert all(x in clns for x in ['CIM_Foo', 'CIM_Foo_sub', 'CIM_Foo_sub2',
                               'CIM_Foo_sub_sub'])
inames = ein('CIM_Foo')
assert len(inames) == 3
assert 'root/cimv2:CIM_Foo.InstanceID="CIM_Foo1"' in [str(n) for n in inames]

# Test get instance for each instance name in inst_nmes
for name in inames:
    inst = gi(name)
    inst_mof = inst.tomof()
    assert inst_mof.startswith('instance of CIM_Foo {')
    print(inst.tomof())
    assert inst.path in inames

quit()
# end of script

"""


# Script that contains a syntax error
TEST_SCRIPT_ERR = 'printx("TEST"); quit()'


QUIT_SCRIPT = """
import sys
# The following is required because wbemcli executes the scripts before
# generating the infomation about connections.  This allows that info to
# be tested as part of stdout return.
from wbemcli import _get_connection_info
print(_get_connection_info())

sys.stdout.flush()
sys.stderr.flush()
quit()

"""

OK = True  # mark tests OK when they execute correctly
RUN = True  # Mark OK = False and current test case being created RUN
FAIL = False  # Any test currently FAILING or not tested yet


TEST_CASES = [
    # desc - String describing the test
    # args - List of arguments. Each argument must be separate member of list
    # exp_response - Dictionary defining expected response. The keywords are:
    #    stdout - Expected response in stdout
    #    stderr - Expected response in stderr
    #    rc - Expected return code. If entry does not exist, 0 is expected
    #    test - Specific test on response.  The allowed tests are defined in
    # mock - Files to be included as attributes of the --mock_server option
    # script - Text for script to be executed by wbemcli with --scripts option
    # condition - Expected test condition

    #
    #   --help
    #
    ['Verify wbemcli help',
     '--help',
     {'stdout': WBEMCLI_HELP,
      'test': 'lines'},
     None, None, OK],

    ['Verify output for log on api=file',
     ['http://blah', '-l', 'api=file'],
     {'stdout': ['log=on', 'Connection: http://blah', ' no creds',
                 'default-namespace=root/cimv2'],
      'test': 'in'},
     None, None, OK],

    ['Verify output for timeout on',
     ['http://blah', '-t', '10'],
     {'stdout': ['timeout=10', 'Connection: http://blah', ' no creds'],
      'test': 'in'},
     None, None, OK],

    ['Verify log definition api=file, connection message',
     ['http://blah', '-l', 'api=file'],
     {'stdout': ['log=on', 'Connection: http://blah', ' no creds'],
      'test': 'in'},
     None, None, OK],

    ['Verify output for log http=file, connection message',
     ['http://blah', '-l', 'http=file'],
     {'stdout': ['log=on', 'Connection: http://blah', ' no creds'],
      'test': 'in'},
     None, None, OK],

    ['Verify output for log all=file, connection message',
     ['http://blah', '-l', 'all=file'],
     {'stdout': ['log=on', 'Connection: http://blah', ' no creds'],
      'test': 'in'},
     None, None, OK],

    ['Verify output for log api=stderr, connection message',
     ['http://blah', '-l', 'api=stderr'],
     {'stdout': ['log=on', 'Connection: http://blah', ' no creds'],
      'test': 'in'},
     None, None, OK],

    ['Validate default default_namespace',
     ['http://localhost'],
     {'stdout': ['Connection: http://localhost', ' no creds',
                 'default-namespace=root/cimv2'],
      'test': 'in'},
     None, None, OK],

    ['Validate user set  default_namespace',
     ['http://localhost', '-n', 'root/blah'],
     {'stdout': ['Connection: http://localhost', ' no creds',
                 'default-namespace=root/blah'],
      'test': 'in'},
     None, None, OK],

    ['Validate load mock mof file',
     [],
     {'stdout': ['Connection: http://FakedUrl', ' no creds',
                 'default-namespace=root/cimv2'],
      'test': 'in'},
     TEST_MOCK_MOF, None, OK],

    ['Validate script load and run. Tests for various operations are in script',
     [],
     {'stdout': ['instance of CIM_Foo', 'InstanceID = "CIM_Foo1";'],
      'test': 'in'},
     TEST_MOCK_MOF, TEST_SCRIPT, OK],

    # TODO: Extend the above to include createinstance
    # TODO: Add invokemethod test. This requires using another mock file
    # TODO: Extend the above to include a mock for associations and refs.
    #       The mock defs for these operations will be in a separate script.
    # TODO: Extend the tests to cover more of the command options (ex.
    #       propertylists, etc.)
    # TODO: We cannot create a test for invalid script because the code
    #       creates the script from a python string in the test definition.
    #       Should find a way to confirm that an invalid file gens error

    #
    #   Test that errors are correctly handled.
    #

    ['Validate invalid url fails',
     ['httpx://localhost'],
     {'stderr': ['error: Invalid scheme on server argument. '
                 'Use "http" or "https"'],
      'rc': 2,
      'test': 'v'},
     None, None, OK],

    ['Validate load invalid mock file fails.',
     ['httpx://localhost', '--mock-server', 'blah.mof'],
     {'stderr': ['error: Build Repository failed: File name blah.mof'],
      'rc': 2,
      'test': 'in'},
     None, None, OK],

    ['Verify invalid timeout param',
     ['http://blah', '-t', 'blah'],
     {'stderr': ["error: argument -t/--timeout: invalid int value: 'blah'"],
      'rc': 2,
      'test': 'in'},
     None, None, OK],

    ['Validate invalid script fails',
     [],
     {'stderr': ["NameError: name 'printx' is not defined"],
      'rc': 1,
      'test': 'in'},
     TEST_MOCK_MOF, TEST_SCRIPT_ERR, OK],
]


@pytest.mark.parametrize(
    "desc, args, exp_response, mock, script, condition",
    TEST_CASES)
def test_command(desc, args, exp_response, mock, script, condition):
    """
    Standard test for pywbemcli commands using pytest
    """
    if not condition:
        pytest.skip("Condition for test case %s not met" % desc)

    script_file_name = create_abs_path("wbemclitestscript.py")

    if script:
        with open(script_file_name, "w") as text_file:
            text_file.write(script)
        args.append("--script")
        args.append(script_file_name)
    else:
        # If not help command, do quit() script to terminate wbemcli
        if not any(s in args for s in ['-h', '--help']):
            with open(script_file_name, "w") as text_file:
                text_file.write(QUIT_SCRIPT)
            args.append("--script")
            args.append(script_file_name)

    try:
        wbemcli_test(desc, args, exp_response, mock, condition, verbose=VERBOSE)

    finally:
        if os.path.isfile(script_file_name):
            os.remove(script_file_name)



#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
*New in pywbem 0.9 as experimental and finalized in 0.10.*

The :class:`~pywbem.WBEMSubscriptionManager` class is a subscription manager
that provides for creating and removing indication subscriptions (including
indication filters and listener destinations) for multiple WBEM servers and
multiple WBEM listeners and for getting information about existing indication
subscriptions.

The WBEM listener is identified through its URL, so it may be the pywbem
listener (that is, a :class:`~pywbem.WBEMListener` object) or any other WBEM
listener.

This subscription manager supports three types of ownership of the CIM
instances in WBEM servers that represent subscriptions, filters and listener
destinations:

* **Owned**

  Owned CIM instances are created via the subscription manager and their life
  cycle is bound to the life cycle of the registration of that WBEM server with
  the subscription manager via
  :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

  Owned CIM instances are deleted automatically when their WBEM server is
  deregistered from the subscription manager via
  :meth:`~pywbem.WBEMSubscriptionManager.remove_server` or
  :meth:`~pywbem.WBEMSubscriptionManager.remove_all_servers`. In addition, they
  can be deleted by the user via the removal methods of the
  :class:`~pywbem.WBEMSubscriptionManager` class.

* **Permanent**

  Permanent CIM instances are created via the subscription manager and their
  life cycle is independent of the life cycle of the registration of that WBEM
  server with the subscription manager.

  Permanent CIM instances are not deleted automatically when their WBEM server
  is deregistered from the subscription manager. The user is responsible for
  their lifetime management: They can be deleted via the removal methods of the
  :class:`~pywbem.WBEMSubscriptionManager` class.

* **Static**

  Static CIM instances pre-exist in the WBEM server and cannot be deleted
  (or created) by a WBEM client.

If a client creates a subscription between a filter and a listener destination,
the types of ownership of these three CIM instances may be arbitrarily mixed,
with one exception:

* A permanent subscription cannot be created on an owned filter or an owned
  listener destination. Allowing that would prevent the automatic life cycle
  management of the owned filter or listener destination by the subscription
  manager. This restriction is enforced by the
  :class:`~pywbem.WBEMSubscriptionManager` class.

The :class:`~pywbem.WBEMSubscriptionManager` object remembers owned
subscriptions, filters, and listener destinations. If for some reason that
object gets deleted before all servers could be removed (e.g. because the
Python program aborts), the corresponding CIM instances in the WBEM server
still exist, but the knowledge is lost that these instances were owned by that
subscription manager. Therefore, the subscription manager discovers owned
subscriptions, filters, and listener destinations when a server is added.
For filters, this discovery is based upon the Name property. Therefore, if
the Name property is set by the user (e.g. because a management profile
requires a particular name), the filter must be permanent and cannot be owned.

Examples
--------

The following example code demonstrates the use of a subscription manager
to subscribe for a CIM alert indication on a WBEM server. The WBEM listener
is assumed to exist somewhere and is identified by its URL::

    import sys
    from socket import getfqdn
    from pywbem import WBEMConnection, WBEMServer, WBEMSubscriptionManager

    server_url = 'http://myserver'
    server_credentials = (user, password)

    listener_url = 'http://mylistener'

    conn = WBEMConnection(server_url, server_credentials)
    server = WBEMServer(conn)
    sub_mgr = WBEMSubscriptionManager(subscription_manager_id='fred')

    # Register the server in the subscription manager:
    server_id = sub_mgr.add_server(server)

    # Add a listener destination in the server:
    dest_inst = sub_mgr.add_listener_destinations(server_id, listener_url)

    # Subscribe to a static filter of a given name:
    filter1_name = "DMTF:Indications:GlobalAlertIndicationFilter"
    filter1_insts = sub_mgr.get_all_filters(server_id)
    for inst in filter1_insts:
        if inst['Name'] == filter1_name:
            sub_mgr.add_subscriptions(server_id, inst.path, dest_inst.path)
            break

    # Create a dynamic alert filter and subscribe to it:
    filter2_source_ns = "root/cimv2"
    filter2_query = "SELECT * FROM CIM_AlertIndication " \\
                    "WHERE OwningEntity = 'DMTF' " \\
                    "AND MessageID LIKE 'SVPC0123|SVPC0124|SVPC0125'"
    filter2_language = "DMTF:CQL"
    filter2_inst = subscription_manager.add_filter(server_id,
        filter2_source_ns, filter2_query, filter2_language, owned=True,
        filter_id="myalert")
    sub_mgr.add_subscriptions(server_id, filter2_inst.path, dest_inst.path)

The following example code briefly shows how to use a subscription manager
as a context manager, in order to get automatic cleanup of owned instances::

    with WBEMSubscriptionManager('fred') as sub_mgr:
        server_id = sub_mgr.add_server(server)
        . . .
    # The exit method automatically calls remove_all_servers(), which deletes
    # all owned subscription, filter and destination instances in the servers
    # that were registered.

The :ref:`Tutorial` section contains a tutorial about the subscription manager.
"""

import re
from socket import getfqdn
import uuid
import six

from ._server import WBEMServer
from .cim_obj import CIMInstance, CIMInstanceName
from .cim_http import parse_url
from .cim_constants import CIM_ERR_FAILED
from .exceptions import CIMError
from ._utils import _format

# CIM model classnames for subscription components
SUBSCRIPTION_CLASSNAME = 'CIM_IndicationSubscription'
DESTINATION_CLASSNAME = 'CIM_ListenerDestinationCIMXML'
FILTER_CLASSNAME = 'CIM_IndicationFilter'
SYSTEM_CREATION_CLASSNAME = 'CIM_ComputerSystem'

DEFAULT_QUERY_LANGUAGE = 'WQL'

__all__ = ['WBEMSubscriptionManager']


class WBEMSubscriptionManager(object):
    """
    *New in pywbem 0.9 as experimental and finalized in 0.10.*

    A class for managing subscriptions for CIM indications in a WBEM server.

    The class may be used as a Python context manager, in order to get
    automatic clean up (see :meth:`~pywbem.WBEMSubscriptionManager.__exit__`).
    """

    def __init__(self, subscription_manager_id):
        """
        Parameters:

          subscription_manager_id (:term:`string`):
            A subscription manager ID string that is used as a component in
            the value of the `Name` property of indication filter and listener
            destination instances to help the user identify these instances in
            a WBEM server.

            Must not be `None`.

            The string must consist of printable characters, and must not
            contain the character ':' because that is the separator between
            components within the value of the `Name` property.

            The subscription manager ID must be unique on the current host.

            For example, the form of the `Name` property of a filter instance
            is (for details, see
            :meth:`~pywbem.WBEMSubscriptionManager.add_filter`):

              ``"pywbemfilter:" {ownership} ":" {subscription_manager_id} ":"
              {filter_id} ":" {guid}``

        Raises:

            ValueError: Incorrect input parameter values.
            TypeError: Incorrect input parameter types.
        """

        # The following dictionaries have the WBEM server ID as a key.
        self._servers = {}  # WBEMServer objects for the WBEM servers
        self._owned_subscriptions = {}  # CIMInstance of owned subscriptions
        self._owned_filters = {}  # CIMInstance of owned filters
        self._owned_destinations = {}  # CIMInstance owned destinations

        if subscription_manager_id is None:
            raise ValueError("Subscription manager ID must not be None")
        if not isinstance(subscription_manager_id, six.string_types):
            raise TypeError(
                _format("Invalid type for subscription manager ID: {0!A}",
                        subscription_manager_id))
        if ':' in subscription_manager_id:
            raise ValueError(
                _format("Subscription manager ID contains ':': {0!A}",
                        subscription_manager_id))
        self._subscription_manager_id = subscription_manager_id

    def __str__(self):
        """
        Return a representation of the :class:`~pywbem.WBEMSubscriptionManager`
        object with a subset of its attributes.
        """
        return _format(
            "WBEMSubscriptionManager("
            "_subscription_manager_id={s._subscription_manager_id!A}, "
            "_servers={s._servers}, "
            "...)",
            s=self)

    def __repr__(self):
        """
        Return a representation of the :class:`~pywbem.WBEMSubscriptionManager`
        object with all attributes, that is suitable for debugging.
        """
        return _format(
            "WBEMSubscriptionManager("
            "_subscription_manager_id={s._subscription_manager_id!A}, "
            "_servers={s._servers!A}, "
            "_owned_subscriptions={s._owned_subscriptions!A}, "
            "_owned_filters={s._owned_filters!A}, "
            "_owned_destinations={s._owned_destinations!A})",
            s=self)

    def __enter__(self):
        """
        *New in pywbem 0.10.*

        Enter method when the class is used as a context manager.
        Returns the subscription manager object
        """
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        *New in pywbem 0.10.*

        Exit method when the class is used as a context manager.

        It cleans up by calling
        :meth:`~pywbem.WBEMSubscriptionManager.remove_all_servers`.
        """
        self.remove_all_servers()
        return False  # re-raise any exceptions

    def _get_server(self, server_id):
        """
        Internal method to get the server object, given a server_id.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

        Returns:

          server (:class:`~pywbem.WBEMServer`):
            The WBEM server.

        Raises:

            ValueError: server_id not known to subscription manager.
        """

        if server_id not in self._servers:
            raise ValueError(
                _format("WBEM server {0!A} not known by subscription manager",
                        server_id))

        return self._servers[server_id]

    def add_server(self, server):
        """
        Register a WBEM server with the subscription manager. This is a
        prerequisite for adding listener destinations, indication filters and
        indication subscriptions to the server.

        This method discovers listener destination, indication filter, and
        subscription instances in the WBEM server owned by this subscription
        manager, and registers them in the subscription manager as if they had
        been created through it.

        In this discovery process, listener destination and indication filter
        instances are matched based upon the value of their `Name` property.
        Subscription instances are matched based upon the ownership of the
        referenced destination and filter instances: If a subscription
        references a filter or a destination that is owned by this subscription
        manager, it is considered owned by this subscription manager as well.

        Parameters:

          server (:class:`~pywbem.WBEMServer`):
            The WBEM server.

        Returns:

            :term:`string`: An ID for the WBEM server, for use by other
            methods of this class.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            ValueError: Incorrect input parameter values.
            TypeError: Incorrect input parameter types.
        """

        if not isinstance(server, WBEMServer):
            raise TypeError("Server argument of add_server() must be a "
                            "WBEMServer object")
        server_id = server.url
        if server_id in self._servers:
            raise ValueError(
                _format("WBEM server already known by listener: {0!A}",
                        server_id))

        # Create dictionary entries for this server
        self._servers[server_id] = server
        self._owned_subscriptions[server_id] = []
        self._owned_filters[server_id] = []
        self._owned_destinations[server_id] = []

        # Recover any owned destination, filter, and subscription instances
        # that exist on this server

        this_host = getfqdn()

        dest_name_pattern = re.compile(
            _format(r'^pywbemdestination:owned:{0}:{1}:[^:]*$',
                    this_host, self._subscription_manager_id))
        dest_insts = server.conn.EnumerateInstances(
            DESTINATION_CLASSNAME, namespace=server.interop_ns)
        for inst in dest_insts:
            if re.match(dest_name_pattern, inst.path.keybindings['Name']) \
                    and inst.path.keybindings['SystemName'] == this_host:
                self._owned_destinations[server_id].append(inst)

        filter_name_pattern = re.compile(
            _format(r'^pywbemfilter:owned:{0}:{1}:[^:]*:[^:]*$',
                    this_host, self._subscription_manager_id))
        filter_insts = server.conn.EnumerateInstances(
            FILTER_CLASSNAME, namespace=server.interop_ns)
        for inst in filter_insts:
            if re.match(filter_name_pattern, inst.path.keybindings['Name']) \
                    and inst.path.keybindings['SystemName'] == this_host:
                self._owned_filters[server_id].append(inst)

        sub_insts = server.conn.EnumerateInstances(
            SUBSCRIPTION_CLASSNAME, namespace=server.interop_ns)
        owned_filter_paths = [inst.path for inst in
                              self._owned_filters[server_id]]
        owned_destination_paths = [inst.path for inst in
                                   self._owned_destinations[server_id]]
        for inst in sub_insts:
            if inst.path.keybindings['Filter'] in \
                    owned_filter_paths \
                    or inst.path.keybindings['Handler'] in \
                    owned_destination_paths:
                self._owned_subscriptions[server_id].append(inst)

        return server_id

    def remove_server(self, server_id):
        """
        Remove a registered WBEM server from the subscription manager. This
        also unregisters listeners from that server and removes all owned
        indication subscriptions, owned indication filters and owned listener
        destinations.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """

        # Validate server_id
        server = self._get_server(server_id)

        # Delete any instances we recorded to be cleaned up

        if server_id in self._owned_subscriptions:
            inst_list = self._owned_subscriptions[server_id]
            # We iterate backwards because we change the list
            for i in six.moves.range(len(inst_list) - 1, -1, -1):
                inst = inst_list[i]
                server.conn.DeleteInstance(inst.path)
                del inst_list[i]
            del self._owned_subscriptions[server_id]

        if server_id in self._owned_filters:
            inst_list = self._owned_filters[server_id]
            # We iterate backwards because we change the list
            for i in six.moves.range(len(inst_list) - 1, -1, -1):
                inst = inst_list[i]
                server.conn.DeleteInstance(inst.path)
                del inst_list[i]
            del self._owned_filters[server_id]

        if server_id in self._owned_destinations:
            inst_list = self._owned_destinations[server_id]
            # We iterate backwards because we change the list
            for i in six.moves.range(len(inst_list) - 1, -1, -1):
                inst = inst_list[i]
                server.conn.DeleteInstance(inst.path)
                del inst_list[i]
            del self._owned_destinations[server_id]

        # Remove server from this listener
        del self._servers[server_id]

    def remove_all_servers(self):
        """
        Remove all registered WBEM servers from the subscription manager. This
        also unregisters listeners from these servers and removes all owned
        indication subscriptions, owned indication filters, and owned listener
        destinations.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """

        for server_id in list(self._servers.keys()):
            self.remove_server(server_id)

    def add_listener_destinations(self, server_id, listener_urls, owned=True):
        """
        Register WBEM listeners to be the target of indications sent by a
        WBEM server.

        This function automatically creates a listener destination instance
        (of CIM class "CIM_ListenerDestinationCIMXML") for each specified
        listener URL in the Interop namespace of the specified WBEM server.

        The form of the `Name` property of the created destination instance is:

          ``"pywbemdestination:" {ownership} ":" {subscription_manager_id} ":"
          {guid}``

        where ``{ownership}`` is ``"owned"`` or ``"permanent"`` dependent on
        the `owned` argument; ``{subscription_manager_id}`` is the
        subscription manager ID; and ``{guid}`` is a globally unique
        identifier.

        Owned listener destinations are added or updated conditionally: If the
        listener destination instance to be added is already registered with
        this subscription manager and has the same property values, it is not
        created or modified. If it has the same path but different property
        values, it is modified to get the desired property values. If an
        instance with this path does not exist yet (the normal case), it is
        created.

        Permanent listener destinations are created unconditionally, and it is
        up to the user to ensure that such an instance does not exist yet.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

          listener_urls (:term:`string` or list of :term:`string`):
            The URL or URLs of the WBEM listeners to be registered.

            The WBEM listener may be a :class:`~pywbem.WBEMListener` object or
            any external WBEM listener.

            Each listener URL string must have the format:

              ``[{scheme}://]{host}:{port}``

            The following URL schemes are supported:

            * ``https``: Causes HTTPS to be used.
            * ``http``: Causes HTTP to be used. This is the default

            The host can be specified in any of the usual formats:

            * a short or fully qualified DNS hostname
            * a literal (= dotted) IPv4 address
            * a literal IPv6 address, formatted as defined in :term:`RFC3986`
              with the extensions for zone identifiers as defined in
              :term:`RFC6874`, supporting ``-`` (minus) for the delimiter
              before the zone ID string, as an additional choice to ``%25``.

            Note that the port is required in listener URLs.

            See :class:`~pywbem.WBEMConnection` for examples of valid URLs,
            with the caveat that the port in server URLs is optional.

          owned (:class:`py:bool`):
            Defines the ownership type of the created listener destination
            instances: If `True`, they will be owned. Otherwise, they will be
            permanent. See :ref:`WBEMSubscriptionManager` for details about
            these ownership types.

        Returns:

            :class:`py:list` of :class:`~pywbem.CIMInstance`: The created
            listener destination instances for the defined listener URLs.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """

        # server_id is validated in _create_...() method.

        # If list, recursively call this function with each list item.
        if isinstance(listener_urls, list):
            dest_insts = []
            for listener_url in listener_urls:
                new_dest_insts = self.add_listener_destinations(
                    server_id, listener_url)
                dest_insts.extend(new_dest_insts)
            return dest_insts

        # Here, the variable will be a single list item.
        listener_url = listener_urls

        dest_inst = self._create_destination(server_id, listener_url, owned)

        return [dest_inst]

    def get_owned_destinations(self, server_id):
        """
        Return the listener destinations in a WBEM server owned by this
        subscription manager.

        This function accesses only the local list of owned destinations; it
        does not contact the WBEM server. The local list of owned destinations
        is discovered from the WBEM server when the server is registered with
        the subscription manager, and is maintained from then on as changes
        happen through this subscription manager.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

        Returns:

            :class:`py:list` of :class:`~pywbem.CIMInstance`: The listener
            destination instances.
        """

        # Validate server_id
        self._get_server(server_id)

        return list(self._owned_destinations[server_id])

    def get_all_destinations(self, server_id):
        """
        Return all listener destinations in a WBEM server.

        This function contacts the WBEM server and retrieves the listener
        destinations by enumerating the instances of CIM class
        "CIM_ListenerDestinationCIMXML" in the Interop namespace of the WBEM
        server.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

        Returns:

            :class:`py:list` of :class:`~pywbem.CIMInstance`: The listener
            destination instances.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """

        # Validate server_id
        server = self._get_server(server_id)

        return server.conn.EnumerateInstances(DESTINATION_CLASSNAME,
                                              namespace=server.interop_ns)

    def remove_destinations(self, server_id, destination_paths):
        # pylint: disable=line-too-long
        """
        Remove listener destinations from a WBEM server, by deleting the
        listener destination instances in the server.

        The listener destinations must be owned or permanent (i.e. not static).

        This method verifies that there are not currently any subscriptions on
        the listener destinations to be removed, in order to handle server
        implementations that do not ensure that on the server side as required
        by :term:`DSP1054`.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

          destination_paths (:class:`~pywbem.CIMInstanceName` or list of :class:`~pywbem.CIMInstanceName`):
            Instance path(s) of the listener destination instance(s) in the
            WBEM server.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_FAILED, if there are referencing subscriptions.
        """  # noqa: E501

        # Validate server_id
        server = self._get_server(server_id)
        conn_id = server.conn.conn_id if server.conn is not None else None

        # If list, recursively call this function with each list item.
        if isinstance(destination_paths, list):
            for dest_path in destination_paths:
                self.remove_destinations(server_id, dest_path)
            return

        # Here, the variable will be a single list item.
        dest_path = destination_paths

        # Verify referencing subscriptions.
        ref_paths = server.conn.ReferenceNames(
            dest_path, ResultClass=SUBSCRIPTION_CLASSNAME)
        if ref_paths:
            # DSP1054 1.2 defines that this CIM error is raised by the server
            # in that case, so we simulate that behavior on the client side.
            raise CIMError(
                CIM_ERR_FAILED,
                "The listener destination is referenced by subscriptions.",
                conn_id=conn_id)

        server.conn.DeleteInstance(dest_path)

        inst_list = self._owned_destinations[server_id]
        # We iterate backwards because we change the list
        for i in six.moves.range(len(inst_list) - 1, -1, -1):
            inst = inst_list[i]
            if inst.path == dest_path:
                del inst_list[i]
                # continue loop to find any possible duplicate entries

    def add_filter(self, server_id, source_namespace, query,
                   query_language=DEFAULT_QUERY_LANGUAGE, owned=True,
                   filter_id=None, name=None):
        """
        Add a :term:`dynamic indication filter` to a WBEM server, by creating
        an indication filter instance (of CIM class "CIM_IndicationFilter") in
        the Interop namespace of the server.

        The `Name` property of the created filter instance can be set in one
        of two ways:

        * directly by specifying the `name` parameter.

          In this case, the `Name` property is set directly to the `name`
          parameter.

          This should be used in cases where the user needs to have control
          over the filter name (e.g. because a DMTF management profile
          requires a particular name), but it cannot be used for owned
          filters.

        * indirectly by specifying the `filter_id` parameter.

          In this case, the value of the `Name` property will be:

          ``"pywbemfilter:" {ownership} ":" {subscription_manager_id} ":"
          {filter_id} ":" {guid}``

          where
          ``{ownership}`` is ``"owned"`` or ``"permanent"`` dependent on
          whether the filter is owned or permmanent;
          ``{subscription_manager_id}`` is the subscription manager ID;
          ``{filter_id}`` is the filter ID; and
          ``{guid}`` is a globally unique identifier.

          This can be used for both owned and permanent filters.

        Owned indication filters are added or updated conditionally: If the
        indication filter instance to be added is already registered with
        this subscription manager and has the same property values, it is not
        created or modified. If it has the same path but different property
        values, it is modified to get the desired property values. If an
        instance with this path does not exist yet (the normal case), it is
        created.

        Permanent indication filters are created unconditionally, and it is
        up to the user to ensure that such an instance does not exist yet.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

          source_namespace (:term:`string`):
            Source namespace of the indication filter.

          query (:term:`string`):
            Filter query in the specified query language.

          query_language (:term:`string`):
            Query language for the specified filter query.

            Examples: 'WQL', 'DMTF:CQL'.

          owned (:class:`py:bool`):
            Defines the ownership type of the created indication filter
            instance: If `True`, it will be owned. Otherwise, it will be
            permanent. See :ref:`WBEMSubscriptionManager` for details about
            these ownership types.

          filter_id (:term:`string`)
            A filter ID string that is used as a component in the value of the
            `Name` property of filter instances to help the user identify
            these instances in a WBEM server, or `None` if the `name`
            parameter is specified.

            The string must consist of printable characters, and must not
            contain the character ':' because that is the separator between
            components within the value of the `Name` property.

            There is no requirement that the filter ID be unique. This can be
            used to identify groups of filters by using the same value for
            multiple filters.

          name (:term:`string`)
            *New in pywbem 0.10.*

            The filter name to be used directly for the `Name` property of the
            filter instance, or `None` if the `filter_id` parameter is
            specified.

        Returns:

            :class:`~pywbem.CIMInstance`: The created indication filter
            instance.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            ValueError: Incorrect input parameter values.
            TypeError: Incorrect input parameter types.
        """

        # server_id is validated in _create_...() method.

        if filter_id is None and name is None:
            raise ValueError("The filter_id and name parameters are both "
                             "None, but exactly one of them must be specified")
        if filter_id is not None:
            if name is not None:
                raise ValueError("The filter_id and name parameters are both "
                                 "specified, but only one of them must be "
                                 "specified")
            if not isinstance(filter_id, six.string_types):
                raise TypeError(
                    _format("Invalid type for filter ID: {0!A}", filter_id))
            if ':' in filter_id:
                raise ValueError(
                    _format("Filter ID contains ':': {0!A}", filter_id))

        filter_inst = self._create_filter(server_id, source_namespace, query,
                                          query_language, owned, filter_id,
                                          name)

        return filter_inst

    def get_owned_filters(self, server_id):
        """
        Return the indication filters in a WBEM server owned by this
        subscription manager.

        This function accesses only the local list of owned filters; it does
        not contact the WBEM server. The local list of owned filters is
        discovered from the WBEM server when the server is registered with the
        the subscription manager, and is maintained from then on as changes
        happen through this subscription manager.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

        Returns:

            :class:`py:list` of :class:`~pywbem.CIMInstance`: The indication
            filter instances.
        """

        # Validate server_id
        self._get_server(server_id)

        return list(self._owned_filters[server_id])

    def get_all_filters(self, server_id):
        """
        Return all indication filters in a WBEM server.

        This function contacts the WBEM server and retrieves the indication
        filters by enumerating the instances of CIM class
        "CIM_IndicationFilter" in the Interop namespace of the WBEM server.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

        Returns:

            :class:`py:list` of :class:`~pywbem.CIMInstance`: The indication
            filter instances.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """

        # Validate server_id
        server = self._get_server(server_id)

        return server.conn.EnumerateInstances('CIM_IndicationFilter',
                                              namespace=server.interop_ns)

    def remove_filter(self, server_id, filter_path):
        """
        Remove an indication filter from a WBEM server, by deleting the
        indication filter instance in the WBEM server.

        The indication filter must be owned or permanent (i.e. not static).

        This method verifies that there are not currently any subscriptions on
        the specified indication filter, in order to handle server
        implementations that do not ensure that on the server side as required
        by :term:`DSP1054`.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

          filter_path (:class:`~pywbem.CIMInstanceName`):
            Instance path of the indication filter instance in the WBEM
            server.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            CIMError: CIM_ERR_FAILED, if there are referencing subscriptions.
        """

        # Validate server_id
        server = self._get_server(server_id)
        conn_id = server.conn.conn_id if server.conn is not None else None

        # Verify referencing subscriptions.
        ref_paths = server.conn.ReferenceNames(
            filter_path, ResultClass=SUBSCRIPTION_CLASSNAME)
        if ref_paths:
            # DSP1054 1.2 defines that this CIM error is raised by the server
            # in that case, so we simulate that behavior on the client side.
            raise CIMError(
                CIM_ERR_FAILED,
                "The indication filter is referenced by subscriptions.",
                conn_id=conn_id)

        server.conn.DeleteInstance(filter_path)

        inst_list = self._owned_filters[server_id]
        # We iterate backwards because we change the list
        for i in six.moves.range(len(inst_list) - 1, -1, -1):
            inst = inst_list[i]
            if inst.path == filter_path:
                del inst_list[i]
                # continue loop to find any possible duplicate entries

    def add_subscriptions(self, server_id, filter_path, destination_paths=None,
                          owned=True):
        # pylint: disable=line-too-long
        """
        Add subscriptions to a WBEM server for a particular set of indications
        defined by an indication filter and for a particular set of WBEM
        listeners defined by the instance paths of their listener destinations,
        by creating indication subscription instances (of CIM class
        "CIM_IndicationSubscription") in the Interop namespace of that server.

        The specified indication filter may be owned, permanent or static.

        The specified listener destinations may be owned, permanent or static.

        When creating permanent subscriptions, the indication filter and the
        listener destinations must not be owned.

        Owned subscriptions are added or updated conditionally: If the
        subscription instance to be added is already registered with
        this subscription manager and has the same property values, it is not
        created or modified. If it has the same path but different property
        values, it is modified to get the desired property values. If an
        instance with this path does not exist yet (the normal case), it is
        created.

        Permanent subscriptions are created unconditionally, and it is up to
        the user to ensure that such an instance does not exist yet.

        Upon successful return of this method, the added subscriptions are
        active, so that the specified WBEM listeners may immediately receive
        indications.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

          filter_path (:class:`~pywbem.CIMInstanceName`):
            Instance path of the indication filter instance in the WBEM
            server that specifies the indications to be sent.

          destination_paths (:class:`~pywbem.CIMInstanceName` or list of :class:`~pywbem.CIMInstanceName`):
            Instance paths of the listener destination instances in the WBEM
            server that specify the target WBEM listener.

            If `None`, subscriptions will be created for all owned listener
            destinations registered to this subscription manager.

          owned (:class:`py:bool`):
            Defines the ownership type of the created subscription
            instances: If `True`, they will be owned. Otherwise, they will be
            permanent. See :ref:`WBEMSubscriptionManager` for details about
            these ownership types.

        Returns:

            :class:`py:list` of :class:`~pywbem.CIMInstance`: The indication
            subscription instances created in the WBEM server.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
            ValueError: Incorrect input parameter values.
        """  # noqa: E501

        # server_id is validated in _create_...() method.

        owned_destination_paths = [inst.path for inst in
                                   self._owned_destinations[server_id]]

        # Apply default
        if destination_paths is None:
            destination_paths = owned_destination_paths

        # If list, recursively call this function with each list item.
        if isinstance(destination_paths, list):
            sub_insts = []
            for dest_path in destination_paths:
                new_sub_insts = self.add_subscriptions(
                    server_id, filter_path, dest_path, owned)
                sub_insts.extend(new_sub_insts)
            return sub_insts

        # Here, the variable will be a single list item.
        dest_path = destination_paths

        owned_filter_paths = [inst.path for inst in
                              self._owned_filters[server_id]]

        # Enforce that a permanent subscription is not created on an owned
        # filter or on an owned destination.
        if not owned:
            if filter_path in owned_filter_paths:
                raise ValueError(
                    _format("Permanent subscription cannot be created on "
                            "owned filter: {0!A}", filter_path))
            if dest_path in owned_destination_paths:
                raise ValueError(
                    _format("Permanent subscription cannot be created on "
                            "owned listener destination: {0!A}", dest_path))

        sub_inst = self._create_subscription(server_id, dest_path, filter_path,
                                             owned)

        return [sub_inst]

    def get_owned_subscriptions(self, server_id):
        """
        Return the indication subscriptions in a WBEM server owned by this
        subscription manager.

        This function accesses only the local list of owned subscriptions; it
        does not contact the WBEM server. The local list of owned subscriptions
        is discovered from the WBEM server when the server is registered with
        the subscription manager, and is maintained from then on as changes
        happen through this subscription manager.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

        Returns:

            :class:`py:list` of :class:`~pywbem.CIMInstance`: The indication
            subscription instances.
        """

        # Validate server_id
        self._get_server(server_id)

        return list(self._owned_subscriptions[server_id])

    def get_all_subscriptions(self, server_id):
        """
        Return all indication subscriptions in a WBEM server.

        This function contacts the WBEM server and retrieves the indication
        subscriptions by enumerating the instances of CIM class
        "CIM_IndicationSubscription" in the Interop namespace of the WBEM
        server.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

        Returns:

            :class:`py:list` of :class:`~pywbem.CIMInstance`: The indication
            subscription instances.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """

        # Validate server_id
        server = self._get_server(server_id)

        return server.conn.EnumerateInstances(SUBSCRIPTION_CLASSNAME,
                                              namespace=server.interop_ns)

    def remove_subscriptions(self, server_id, sub_paths):
        # pylint: disable=line-too-long
        """
        Remove indication subscription(s) from a WBEM server, by deleting the
        indication subscription instances in the server.

        The indication subscriptions must be owned or permanent (i.e. not
        static).

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

          sub_paths (:class:`~pywbem.CIMInstanceName` or list of :class:`~pywbem.CIMInstanceName`):
            Instance path(s) of the indication subscription instance(s) in the
            WBEM server.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """  # noqa: E501

        # Validate server_id
        server = self._get_server(server_id)

        # If list, recursively call this function with each list item.
        if isinstance(sub_paths, list):
            for sub_path in sub_paths:
                self.remove_subscriptions(server_id, sub_path)
            return

        # Here, the variable will be a single list item.
        sub_path = sub_paths

        server.conn.DeleteInstance(sub_path)

        inst_list = self._owned_subscriptions[server_id]
        # We iterate backwards because we change the list
        for i in six.moves.range(len(inst_list) - 1, -1, -1):
            inst = inst_list[i]
            if inst.path == sub_path:
                del inst_list[i]
                # continue loop to find any possible duplicate entries

    def _create_destination(self, server_id, dest_url, owned):
        """
        Create a listener destination instance in the Interop namespace of a
        WBEM server and return that instance.

        In order to catch any changes the server applies, the instance is
        retrieved again using the instance path returned by instance creation.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

          dest_url (:term:`string`):
            URL of the listener that is used by the WBEM server to send any
            indications to.

            The URL scheme (e.g. http/https) determines whether the WBEM server
            uses HTTP or HTTPS for sending the indication. Host and port in the
            URL specify the target location to be used by the WBEM server.

          owned (:class:`py:bool`):
            Defines whether or not the created instance is *owned* by the
            subscription manager.

        Returns:

            :class:`~pywbem.CIMInstance`: The created instance, as retrieved
            from the server.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """

        # Validate server_id
        server = self._get_server(server_id)

        # validate the URL by reconstructing it. Do not allow defaults
        host, port, ssl = parse_url(dest_url, allow_defaults=False)
        schema = 'https' if ssl else 'http'
        listener_url = '{0}://{1}:{2}'.format(schema, host, port)

        this_host = getfqdn()
        ownership = "owned" if owned else "permanent"

        dest_path = CIMInstanceName(DESTINATION_CLASSNAME,
                                    namespace=server.interop_ns)

        dest_inst = CIMInstance(DESTINATION_CLASSNAME)
        dest_inst.path = dest_path
        dest_inst['CreationClassName'] = DESTINATION_CLASSNAME
        dest_inst['SystemCreationClassName'] = SYSTEM_CREATION_CLASSNAME
        dest_inst['SystemName'] = this_host
        dest_inst['Name'] = _format(
            'pywbemdestination:{0}:{1}:{2}',
            ownership, self._subscription_manager_id, uuid.uuid4())
        dest_inst['Destination'] = listener_url

        if owned:
            for i, inst in enumerate(self._owned_destinations[server_id]):
                if inst.path == dest_path:
                    # It already exists, now check its properties
                    if inst != dest_inst:
                        server.conn.ModifyInstance(dest_inst)
                        dest_inst = server.conn.GetInstance(dest_path)
                        self._owned_destinations[server_id][i] = dest_inst
                    return dest_inst
            dest_path = server.conn.CreateInstance(dest_inst)
            dest_inst = server.conn.GetInstance(dest_path)
            self._owned_destinations[server_id].append(dest_inst)
            return dest_inst
        else:
            # Responsibility to ensure it does not exist yet is with the user
            dest_path = server.conn.CreateInstance(dest_inst)
            dest_inst = server.conn.GetInstance(dest_path)
            return dest_inst

    def _create_filter(self, server_id, source_namespace, query,
                       query_language, owned, filter_id, name):
        """
        Create a :term:`dynamic indication filter` instance in the Interop
        namespace of a WBEM server and return that instance.

        In order to catch any changes the server applies, the instance is
        retrieved again using the instance path returned by instance creation.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

          source_namespace (:term:`string`):
            Source namespace of the indication filter.

          query (:term:`string`):
            Filter query in the specified query language.

          query_language (:term:`string`):
            Query language for the specified filter query.

            Examples: 'WQL', 'DMTF:CQL'.

          owned (:class:`py:bool`):
            Defines whether or not the created instance is *owned* by the
            subscription manager.

          filter_id (:term:`string`):
            Filter ID string to be incorporated into the `Name` property of the
            filter instance, as detailed for `subscription_manager_id`, or
            `None`.
            Mutually exclusive with the `name` parameter.

          name (:term:`string`):
            Value for the `Name` property of the filter instance, or `None`.
            Mutually exclusive with the `filter_id` parameter.

        Returns:

            :class:`~pywbem.CIMInstance`: The created instance, as retrieved
            from the server.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """

        # Validate server_id
        server = self._get_server(server_id)

        this_host = getfqdn()
        ownership = "owned" if owned else "permanent"

        filter_path = CIMInstanceName(FILTER_CLASSNAME,
                                      namespace=server.interop_ns)

        filter_inst = CIMInstance(FILTER_CLASSNAME)
        filter_inst.path = filter_path
        filter_inst['CreationClassName'] = FILTER_CLASSNAME
        filter_inst['SystemCreationClassName'] = SYSTEM_CREATION_CLASSNAME
        filter_inst['SystemName'] = this_host
        if filter_id:
            filter_inst['Name'] = _format(
                'pywbemfilter:{0}:{1}:{2}:{3}',
                ownership, self._subscription_manager_id, filter_id,
                uuid.uuid4())
        if name:
            filter_inst['Name'] = name
        filter_inst['SourceNamespace'] = source_namespace
        filter_inst['Query'] = query
        filter_inst['QueryLanguage'] = query_language

        if owned:
            for i, inst in enumerate(self._owned_filters[server_id]):
                if inst.path == filter_path:
                    # It already exists, now check its properties
                    if inst != filter_inst:
                        server.conn.ModifyInstance(filter_inst)
                        filter_inst = server.conn.GetInstance(filter_path)
                        self._owned_filters[server_id][i] = filter_inst
                    return filter_inst
            filter_path = server.conn.CreateInstance(filter_inst)
            filter_inst = server.conn.GetInstance(filter_path)
            self._owned_filters[server_id].append(filter_inst)
            return filter_inst
        else:
            # Responsibility to ensure it does not exist yet is with the user
            filter_path = server.conn.CreateInstance(filter_inst)
            filter_inst = server.conn.GetInstance(filter_path)
            return filter_inst

    def _create_subscription(self, server_id, dest_path, filter_path, owned):
        """
        Create an indication subscription instance in the Interop namespace of
        a WBEM server and return that instance.

        In order to catch any changes the server applies, the instance is
        retrieved again using the instance path returned by instance creation.

        Parameters:

          server_id (:term:`string`):
            The server ID of the WBEM server, returned by
            :meth:`~pywbem.WBEMSubscriptionManager.add_server`.

          dest_path (:class:`~pywbem.CIMInstanceName`):
            Instance path of the listener destination instance in the WBEM
            server that references this listener.

          filter_path (:class:`~pywbem.CIMInstanceName`):
            Instance path of the indication filter instance in the WBEM
            server that specifies the indications to be sent.

          owned (:class:`py:bool`):
            Defines whether or not the created instance is *owned* by the
            subscription manager.

        Returns:

            :class:`~pywbem.CIMInstance`: The created instance, as retrieved
            from the server.

        Raises:

            Exceptions raised by :class:`~pywbem.WBEMConnection`.
        """

        # Validate server_id
        server = self._get_server(server_id)

        sub_path = CIMInstanceName(SUBSCRIPTION_CLASSNAME,
                                   namespace=server.interop_ns)

        sub_inst = CIMInstance(SUBSCRIPTION_CLASSNAME)
        sub_inst.path = sub_path
        sub_inst['Filter'] = filter_path
        sub_inst['Handler'] = dest_path

        if owned:
            for inst in self._owned_subscriptions[server_id]:
                if inst.path == sub_path:
                    # It does not have any properties besides its keys,
                    # so checking the path is sufficient.
                    return sub_inst
            sub_path = server.conn.CreateInstance(sub_inst)
            sub_inst = server.conn.GetInstance(sub_path)
            self._owned_subscriptions[server_id].append(sub_inst)
            return sub_inst
        else:
            # Responsibility to ensure it does not exist yet is with the user
            sub_path = server.conn.CreateInstance(sub_inst)
            sub_inst = server.conn.GetInstance(sub_path)
            return sub_inst



"""
Pywbem scriptlet to quit wbemcli. It is used
by test_wbemcli.py to force wbemcli to terminate after each test.
"""

from __future__ import absolute_import, print_function

# This script displays the connection info.  Required because what you see in
# wbemcli startup is through the python interactive module and not to
# stdout
# pylint: disable=undefined-variable
print(_get_connection_info())  # noqa: F821
quit()




#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#

"""
**Experimental:** *New in pywbem 0.9 as experimental.*

Operation recorder interface and implementations.
"""

from __future__ import print_function, absolute_import
from collections import namedtuple
try:
    from collections import OrderedDict
except ImportError:
    from ordereddict import OrderedDict  # pylint: disable=import-error
from datetime import datetime, timedelta
import logging

import six
import yaml
from yaml.representer import RepresenterError

from ._nocasedict import NocaseDict
from .cim_obj import CIMInstance, CIMInstanceName, CIMClass, CIMClassName, \
    CIMProperty, CIMMethod, CIMParameter, CIMQualifier, \
    CIMQualifierDeclaration
from .cim_types import CIMInt, CIMFloat, CIMDateTime
from .exceptions import CIMError
from ._logging import LOGGER_API_CALLS_NAME, LOGGER_HTTP_NAME
from ._utils import _ensure_unicode, _format

if six.PY2:
    import codecs  # pylint: disable=wrong-import-order


__all__ = ['BaseOperationRecorder', 'TestClientRecorder',
           'LogOperationRecorder',
           'OpArgs', 'OpResult', 'HttpRequest', 'HttpResponse']

if six.PY2:
    _Longint = long  # noqa: F821
else:
    _Longint = int

OpArgsTuple = namedtuple("OpArgsTuple", ["method", "args"])


def _represent_ordereddict(dump, tag, mapping, flow_style=None):
    """PyYAML representer function for OrderedDict.
    This is needed for yaml.safe_dump() to support OrderedDict.

    Courtesy:
    https://blog.elsdoerfer.name/2012/07/26/make-pyyaml-output-an-ordereddict/
    """
    value = []
    node = yaml.MappingNode(tag, value, flow_style=flow_style)
    if dump.alias_key is not None:
        dump.represented_objects[dump.alias_key] = node
    best_style = True
    if hasattr(mapping, 'items'):
        mapping = mapping.items()
    for item_key, item_value in mapping:
        node_key = dump.represent_data(item_key)
        node_value = dump.represent_data(item_value)
        if not (isinstance(node_key, yaml.ScalarNode) and
                not node_key.style):
            best_style = False    # pylint: disable=bad-indentation
        if not (isinstance(node_value, yaml.ScalarNode) and
                not node_value.style):
            best_style = False    # pylint: disable=bad-indentation
        value.append((node_key, node_value))
    if flow_style is None:
        if dump.default_flow_style is not None:
            node.flow_style = dump.default_flow_style
        else:
            node.flow_style = best_style
    return node


yaml.SafeDumper.add_representer(
    OrderedDict,
    lambda dumper, value:
    _represent_ordereddict(dumper, u'tag:yaml.org,2002:map', value))


# Tag for CIMDateTime serialization in yaml files
CIMDATETIME_TAG = '!CIMDateTime'


def _cimdatetime_representer(dumper, cimdatetime):
    """
    PyYAML representer function for CIMDateTime objects.
    This is needed for yaml.safe_dump() to support CIMDateTime.
    """
    cimdatetime_str = str(cimdatetime)
    node = dumper.represent_scalar(CIMDATETIME_TAG, cimdatetime_str)
    return node


yaml.SafeDumper.add_representer(CIMDateTime, _cimdatetime_representer)


def _cimdatetime_constructor(loader, node):
    """
    PyYAML constructor function for CIMDateTime objects.
    This is needed for yaml.safe_load() to support CIMDateTime.
    """
    cimdatetime_str = loader.construct_scalar(node)
    cimdatetime = CIMDateTime(cimdatetime_str)
    return cimdatetime


yaml.SafeLoader.add_constructor(CIMDATETIME_TAG, _cimdatetime_constructor)


# Some monkey-patching for better diagnostics:
def _represent_undefined(self, data):
    """Raises flag for objects that cannot be represented"""
    raise RepresenterError(
        _format("Cannot represent an object: {0!A} of type: {1}; "
                "yaml_representers: {2!A}, "
                "yaml_multi_representers: {3!A}",
                data, type(data), self.yaml_representers.keys(),
                self.yaml_multi_representers.keys()))


yaml.SafeDumper.represent_undefined = _represent_undefined


class OpArgs(OpArgsTuple):
    """
    A named tuple representing the name and input arguments of the invocation
    of a :class:`~pywbem.WBEMConnection` method.

    **Experimental:** *New in pywbem 0.9 as experimental.*

    Attributes:

      ~OpArgs.method (:term:`unicode string`):
        Name of the :class:`~pywbem.WBEMConnection` method.

      ~OpArgs.args (:class:`py:dict`):
        Dictionary of input arguments (both positional and keyword-based).
    """
    __slots__ = ()

    def __repr__(self):
        return _format(
            "OpArgs("
            "method={s.method!A}, "
            "args={s.args!A})",
            s=self)


OpResultTuple = namedtuple("OpResultTuple", ["ret", "exc"])


class OpResult(OpResultTuple):
    """
    A named tuple representing the result of the invocation of a
    :class:`~pywbem.WBEMConnection` method.

    **Experimental:** *New in pywbem 0.9 as experimental.*

    Attributes:

      ~OpResult.ret (:class:`py:object`):
        Return value, if the method returned.
        `None`, if the method raised an exception.

        Note that `None` may be a legitimate return value, so the test for
        exceptions should be done based upon the :attr:`exc` variable.

      ~OpResult.exc (:exc:`~py:exceptions.Exception`):
        Exception object, if the method raised an exception.
        `None`, if the method returned.
    """
    __slots__ = ()

    def __repr__(self):
        return _format(
            "OpResult("
            "ret={s.ret!A}, "
            "exc={s.exc!A})",
            s=self)


HttpRequestTuple = namedtuple("HttpRequestTuple",
                              ["version", "url", "target", "method", "headers",
                               "payload"])


class HttpRequest(HttpRequestTuple):
    """
    A named tuple representing the HTTP request sent by the WBEM client.

    **Experimental:** *New in pywbem 0.9 as experimental.*

    Attributes:

      ~HttpRequest.version (:term:`number`):
        HTTP version from the request line (10 for HTTP/1.0, 11 for HTTP/1.1).

      ~HttpRequest.url (:term:`unicode string`):
        URL of the WBEM server (e.g. 'https://myserver.acme.com:15989').

      ~HttpRequest.target (:term:`unicode string`):
        Target URL segment as stated in request line (e.g. '/cimom').

      ~HttpRequest.method (:term:`unicode string`):
        HTTP method as stated in the request line (e.g. "POST").

      ~HttpRequest.headers (:class:`py:dict`):
        A dictionary of all HTTP header fields:

        * key (:term:`unicode string`): Name of the header field
        * value (:term:`unicode string`): Value of the header field

      ~HttpRequest.payload (:term:`unicode string`):
        HTTP payload, i.e. the CIM-XML string.
    """
    __slots__ = ()

    def __repr__(self):
        return _format(
            "HttpRequest("
            "version={s.version!A}, "
            "url={s.url!A}, "
            "target={s.target!A}, "
            "method={s.method!A}, "
            "headers={s.headers!A}, "
            "payload={s.payload!A})",
            s=self)


HttpResponseTuple = namedtuple("HttpResponseTuple",
                               ["version", "status", "reason", "headers",
                                "payload"])


class HttpResponse(HttpResponseTuple):
    """
    A named tuple representing the HTTP response received by the WBEM client.

    **Experimental:** *New in pywbem 0.9 as experimental.*

    Attributes:

      ~HttpResponse.version (:term:`number`):
        HTTP version from the response line (10 for HTTP/1.0, 11 for HTTP/1.1).

      ~HttpResponse.status (:term:`number`):
        HTTP status code from the response line (e.g. 200).

      ~HttpResponse.reason (:term:`unicode string`):
        HTTP reason phrase from the response line (e.g. "OK").

      ~HttpResponse.headers (:class:`py:dict`):
        A dictionary of all HTTP header fields:

        * key (:term:`unicode string`): Name of the header field
        * value (:term:`unicode string`): Value of the header field

      ~HttpResponse.payload (:term:`unicode string`):
        HTTP payload, i.e. the CIM-XML string.
    """
    __slots__ = ()

    def __repr__(self):
        return _format(
            "HttpResponse("
            "version={s.version!A}, "
            "status={s.status!A}, "
            "reason={s.reason!A}, "
            "headers={s.headers!A}, "
            "payload={s.payload!A})",
            s=self)


class BaseOperationRecorder(object):
    # pylint: disable=too-many-instance-attributes
    """
    Abstract base class defining the interface to an operation recorder,
    that records the WBEM operations executed in a connection to a WBEM
    server.

    **Experimental:** *New in pywbem 0.9 as experimental.*

    An operation recorder can be added to a connection via the
    :meth:`~pywbem.WBEMConnection.add_operation_recorder` method. The operation
    recorders of a connection can be retrieved via the
    :attr:`~pywbem.WBEMConnection.operation_recorders` property.

    Each operation that is executed on a connection will cause the
    :meth:`record` method of those operation recorders of the connection to be
    called, that are enabled.

    After being added to a connection, an operation recorder is enabled. It can
    be disabled and re-enabled using the
    :meth:`~pywbem.BaseOperationRecorder.disable` and
    :meth:`~pywbem.BaseOperationRecorder.enable` methods, respectively.
    This can be used to temporarily pause the recorder.
    """

    def __init__(self):
        self._enabled = True
        self._conn_id = None
        self.reset()

    def enable(self):
        """
        Enable the recorder.

        *New in pywbem 0.10.*
        """
        self._enabled = True

    def disable(self):
        """
        Disable the recorder.

        *New in pywbem 0.10.*
        """
        self._enabled = False

    @property
    def enabled(self):
        """
        Indicate whether the recorder is enabled.

        *New in pywbem 0.10.*
        """
        return self._enabled

    @staticmethod
    def open_file(filename, file_mode='w'):
        """
        A static convenience function that performs the open of the recorder
        file correctly for different versions of Python.

        *New in pywbem 0.10.*

        This covers the issue where the file should be opened in text mode but
        that is done differently in Python 2 and Python 3.

        The returned file-like object must be closed by the caller.

        Parameters:

          filename(:term:`string`):
            Name of the file where the recorder output will be written

          file_mode(:term:`string`):
            Optional file mode.  The default is 'w' which overwrites any
            existing file.  if 'a' is used, the data is appended to any
            existing file.

        Returns:

          File-like object.

        Example::

            recorder = TestClientRecorder(...)
            recorder_file = recorder.open_file('recorder.log')

            . . . # Perform WBEM operations using the recorder

            recorder_file.close()
        """
        if six.PY2:
            # Open with codecs to define text mode
            return codecs.open(filename, mode=file_mode, encoding='utf-8')

        return open(filename, file_mode, encoding='utf8')

    def reset(self, pull_op=None):
        """Reset all the attributes in the class. This also allows setting
        the pull_op attribute that defines whether the operation is to be
        a traditional or pull operation.
        This does NOT reset _conn.id as that exists through the life of
        the connection.
        """
        self._pywbem_method = None
        self._pywbem_args = None

        self._pywbem_result_ret = None
        self._pywbem_result_exc = None

        self._http_request_version = None
        self._http_request_url = None
        self._http_request_target = None
        self._http_request_method = None
        self._http_request_headers = None
        self._http_request_payload = None

        self._http_response_version = None
        self._http_response_status = None
        self._http_response_reason = None
        self._http_response_headers = None
        self._http_response_payload = None
        self._pull_op = pull_op

    def stage_wbem_connection(self, wbem_connection):
        """
        Stage information about the connection. Used only by
        LogOperationRecorder.
        """
        pass

    def stage_pywbem_args(self, method, **kwargs):
        """
        Set requst method and all args.
        Normally called before the cmd is executed to record request
        parameters
        """
        # pylint: disable=attribute-defined-outside-init
        self._pywbem_method = method
        self._pywbem_args = kwargs

    def stage_pywbem_result(self, ret, exc):
        """ Set Result return info or exception info"""
        # pylint: disable=attribute-defined-outside-init
        self._pywbem_result_ret = ret
        self._pywbem_result_exc = exc

    def stage_http_request(self, conn_id, version, url, target, method, headers,
                           payload):
        """Set request HTTP information including url, headers, etc."""
        # pylint: disable=attribute-defined-outside-init
        self._http_request_version = version
        self._http_request_conn_id = conn_id
        self._http_request_url = url
        self._http_request_target = target
        self._http_request_method = method
        self._http_request_headers = headers
        self._http_request_payload = payload

    # pylint: disable=unused-argument
    def stage_http_response1(self, conn_id, version, status, reason, headers):
        """Set response http info including headers, status, etc.
           conn_id unused here. Used in log"""
        # pylint: disable=attribute-defined-outside-init
        self._http_response_version = version
        self._http_response_status = status
        self._http_response_reason = reason
        self._http_response_headers = headers

    def stage_http_response2(self, payload):
        """Stage second part of http response, the payload"""
        # pylint: disable=attribute-defined-outside-init
        self._http_response_payload = payload

    def record_staged(self):
        """Encode staged information on request and result to output"""
        if self.enabled:
            pwargs = OpArgs(
                self._pywbem_method,
                self._pywbem_args)
            pwresult = OpResult(
                self._pywbem_result_ret,
                self._pywbem_result_exc)
            httpreq = HttpRequest(
                self._http_request_version,
                self._http_request_url,
                self._http_request_target,
                self._http_request_method,
                self._http_request_headers,
                self._http_request_payload)
            httpresp = HttpResponse(
                self._http_response_version,
                self._http_response_status,
                self._http_response_reason,
                self._http_response_headers,
                self._http_response_payload)
            self.record(pwargs, pwresult, httpreq, httpresp)

    def record(self, pywbem_args, pywbem_result, http_request, http_response):
        """
        Function that is called to record a single WBEM operation, i.e. the
        invocation of a single :class:`~pywbem.WBEMConnection` method.

        This function is called only when the recorder is enabled, i.e. it
        does not need to check for recorder enablement.

        Parameters:

          pywbem_args (:class:`~pywbem.OpArgs`):
            The name and input arguments of the :class:`~pywbem.WBEMConnection`
            method that is recorded.

          pywbem_result (:class:`~pywbem.OpResult`):
            The result (return value or exception) of the
            :class:`~pywbem.WBEMConnection` method that is recorded.

          http_request (:class:`~pywbem.HttpRequest`):
            The HTTP request sent by the :class:`~pywbem.WBEMConnection` method
            that is recorded.

            `None`, if no HTTP request had been sent (e.g. because an exception
            was raised before getting there).

          http_response (:class:`~pywbem.HttpResponse`):
            The HTTP response received by the :class:`~pywbem.WBEMConnection`
            method that is recorded.

            `None`, if no HTTP response had been received (e.g. because an
            exception was raised before getting there).
        """
        raise NotImplementedError


class LogOperationRecorder(BaseOperationRecorder):
    """
    A recorder that logs certain aspects of the WBEM operations driven by
    pywbem users to Python loggers.

    **Experimental:** *New in pywbem 0.11 and redesigned in pywbem 0.12 as
    experimental.*

    This recorder supports two Python loggers:

    * API logger (Python logger name: `'pywbem.api'`) - Logs API calls and
      returns, for the :class:`~pywbem.WBEMConnection` methods that drive WBEM
      operations (see :ref:`WBEM operations`). This logs the API parameters and
      results, including CIM objects/exceptions. It also logs the creation of
      :class:`~pywbem.WBEMConnection` objects to capture connection information
      in order to determine the connection to which a particular log record
      belongs.

    * HTTP logger (Python logger name: `'pywbem.http'`) - Logs HTTP requests
      and responses between the pywbem client and WBEM server. This logs the
      HTTP request data and response data including HTTP headers and CIM-XML
      payload.

    All logging calls are at the :attr:`py:logging.DEBUG` logging level.
    """
    def __init__(self, conn_id, detail_levels=None):
        """
        Parameters:

          conn_id (:term:`connection id`):
            String that represents an id for a particular connection
            that is used to build the name of each logger. The logger names
            are uniqueqly idenfified by the conn_id suffix to the logger name
            (ex.pywbem.ops.1-2343) so that each WBEMConnection could be
            logged with a separate logger.

          detail_levels (:class:`py:dict`):
            Dictionary identifying the detail level for one or both of
            the loggers.
            Key: Simple logger name (e.g. 'api').
            Value: Detail level, either a string from
            :data:`~pywbem._logging.LOG_DETAIL_LEVELS`, or an integer that
            specifies the maximum size of each log record.
        """
        super(LogOperationRecorder, self).__init__()

        self._conn_id = conn_id

        self.detail_levels = {}

        # logging only occurs if the corresponding detail level is not None
        self.api_detail_level = None
        self.http_detail_level = None
        self.api_maxlen = None
        self.http_maxlen = None
        self.set_detail_level(detail_levels)

        # build name for logger for this connection
        if conn_id:
            self.apilogger = logging.getLogger(
                _format("{0}.{1}", LOGGER_API_CALLS_NAME, conn_id))
            self.httplogger = logging.getLogger(
                _format("{0}.{1}", LOGGER_HTTP_NAME, conn_id))
        else:
            self.apilogger = logging.getLogger(LOGGER_API_CALLS_NAME)
            self.httplogger = logging.getLogger(LOGGER_HTTP_NAME)

    def set_detail_level(self, detail_levels):
        """
        Sets the detail levels from the input dictionary in detail_levels.
        """
        if detail_levels is None:
            return

        self.detail_levels = detail_levels
        if 'api' in detail_levels:
            self.api_detail_level = detail_levels['api']
        if 'http' in detail_levels:
            self.http_detail_level = detail_levels['http']
        if isinstance(self.api_detail_level, int):
            self.api_maxlen = self.api_detail_level
        if isinstance(self.http_detail_level, int):
            self.http_maxlen = self.http_detail_level

    def stage_wbem_connection(self, wbem_connection):
        """
        Log connection information. This includes the connection id (conn_id)
        that is output with the log entry. This entry is logged if either
        http or api loggers are enable. It honors both the logger and
        detail level of either api logger if defined or http logger if defined.
        If the api logger does not exist, the output shows this as an http
        loggger output since we do not want to create an api logger for this
        specific output
        """
        self._conn_id = wbem_connection.conn_id

        if self.enabled:
            if self.api_detail_level is not None:
                logger = self.apilogger
                detail_level = self.api_detail_level
                max_len = self.api_maxlen
            elif self.http_detail_level is not None:
                logger = self.httplogger
                detail_level = self.http_detail_level
                max_len = self.http_maxlen
            else:
                return

            if logger.isEnabledFor(logging.DEBUG):
                conn_data = str(wbem_connection) if detail_level == 'summary' \
                    else repr(wbem_connection)

                if max_len and (len(conn_data) > max_len):
                    conn_data = conn_data[:max_len] + '...'
                logger.debug('Connection:%s %s', self._conn_id, conn_data)

    def stage_pywbem_args(self, method, **kwargs):
        """
        Log request method and all args.
        Normally called before the cmd is executed to record request
        parameters.
        This method does not support the summary detail_level because
        that seems to add little info to the log that is not also in the
        response.
        """
        # pylint: disable=attribute-defined-outside-init
        self._pywbem_method = method
        if self.enabled and self.api_detail_level is not None and \
                self.apilogger.isEnabledFor(logging.DEBUG):

            # TODO: future bypassed code to only ouput name and method if the
            # detail is summary.  We are not doing this because this is
            # effectively the same information in the response so the only
            # additional infomation is the time stamp.

            # if self.api_detail_level == summary:
            #    self.apilogger.debug('Request:%s %s', self._conn_id, method)
            #    return

            # Order kwargs.  Note that this is done automatically starting
            # with python 3.6
            kwstr = ', '.join([('{0}={1!r}'.format(key, kwargs[key]))
                               for key in sorted(six.iterkeys(kwargs))])

            if self.api_maxlen and (len(kwstr) > self.api_maxlen):
                kwstr = kwstr[:self.api_maxlen] + '...'
            # pylint: disable=bad-continuation
            self.apilogger.debug('Request:%s %s(%s)', self._conn_id, method,
                                 kwstr)

    def stage_pywbem_result(self, ret, exc):
        """
        Log result return or exception parameter. This method provides varied
        type of formatting based on the detail_level parameter and the
        data in ret.
        """
        def format_result(ret, max_len):
            """ format ret as repr while clipping it to max_len if
                max_len is not None.
            """
            # format the 'summary' and 'paths' detail_levels
            if self.api_detail_level == 'summary':
                if isinstance(ret, list):
                    if ret:
                        ret_type = type(ret[0]).__name__ if ret else ""
                        return _format("list of {0}; count={1}",
                                       ret_type, len(ret))
                    return "Empty"

                ret_type = type(ret).__name__
                if hasattr(ret, 'classname'):
                    name = ret.classname
                elif hasattr(ret, 'name'):
                    name = ret.name
                else:
                    name = ""
                return _format("{0} {1}", ret_type, name)

            elif self.api_detail_level == 'paths':
                if isinstance(ret, list):
                    if ret:
                        if hasattr(ret[0], 'path'):
                            result_fmt = ', '.join(
                                [_format("{0!A}", str(p.path)) for p in ret])
                        else:
                            result_fmt = _format("{0!A}", ret)
                    else:
                        result_fmt = ''
                else:
                    if hasattr(ret, 'path'):
                        result_fmt = _format("{0!A}", str(ret.path))
                    else:
                        result_fmt = _format("{0!A}", ret)

            else:
                result_fmt = _format("{0!A}", ret)

            if max_len and (len(result_fmt) > max_len):
                result_fmt = result_fmt[:max_len] + '...'
            return result_fmt

        if self.enabled and self.api_detail_level is not None and \
                self.apilogger.isEnabledFor(logging.DEBUG):
            if exc:  # format exception
                # exceptions are always either all or reduced length
                result = format_result(
                    _format("{0}({1})", exc.__class__.__name__, exc),
                    self.api_maxlen)

                return_type = 'Exception'

            else:    # format result
                # test if type is tuple (subclass of tuple but not type tuple)
                # pylint: disable=unidiomatic-typecheck
                qrc = ""
                # format open/pull response
                if isinstance(ret, tuple) and \
                        type(ret) is not tuple:  # pylint: disable=C0123
                    try:    # test if field instances or paths
                        rtn_data = ret.instances
                        data_str = 'instances'
                    except AttributeError:
                        rtn_data = ret.paths
                        data_str = 'paths'

                    try:    # test for query_result_class
                        qrc = _format(
                            ", query_result_class={0}",
                            ret.query_result_class)
                    except AttributeError:
                        pass

                    result = _format(
                        "{0}(context={1}, eos={2}{3}, {4}={5})",
                        type(ret).__name__, ret.context, ret.eos, qrc,
                        data_str, format_result(rtn_data, self.api_maxlen))

                # format enumerate response except not open/pull
                elif isinstance(ret, list):
                    try:    # test for query_result_class
                        qrc = _format(
                            ", query_result_class={0}",
                            ret.query_result_class)
                    except AttributeError:
                        pass
                    ret_fmtd = format_result(ret, self.api_maxlen)
                    result = _format("{0}{1}", qrc, ret_fmtd)

                # format single return object
                else:
                    result = format_result(ret, self.api_maxlen)

                return_type = 'Return'

            self.apilogger.debug('%s:%s %s(%s)', return_type, self._conn_id,
                                 self._pywbem_method, result)

    def stage_http_request(self, conn_id, version, url, target, method, headers,
                           payload):
        """Log request HTTP information including url, headers, etc."""
        if self.enabled and self.http_detail_level is not None and \
                self.httplogger.isEnabledFor(logging.DEBUG):
            # pylint: disable=attribute-defined-outside-init
            # if Auth header, mask data
            if 'Authorization' in headers:
                authtype, cred = headers['Authorization'].split(' ')
                headers['Authorization'] = _format(
                    "{0} {1}", authtype, 'X' * len(cred))

            header_str = ' '.join('{0}:{1!r}'.format(k, v)
                                  for k, v in headers.items())
            if self.http_detail_level == 'summary':
                upayload = ""
            elif isinstance(payload, six.binary_type):
                upayload = payload.decode('utf-8')
            else:
                upayload = payload
            if self.http_maxlen and (len(payload) > self.http_maxlen):
                upayload = upayload[:self.http_maxlen] + '...'

            self.httplogger.debug('Request:%s %s %s %s %s %s\n    %s',
                                  conn_id, method, target, version, url,
                                  header_str, upayload)

    def stage_http_response1(self, conn_id, version, status, reason, headers):
        """Set response http info including headers, status, etc. """
        # pylint: disable=attribute-defined-outside-init
        self._http_response_version = version
        self._http_response_status = status
        self._http_response_reason = reason
        self._http_response_headers = headers
        self._http_response_conn_id = conn_id

    def stage_http_response2(self, payload):
        """Log complete http response, including response1 and payload"""

        # required because http code uses sending all None to reset
        # parameters. We ignore that
        if not self._http_response_version and not payload:
            return
        if self.enabled and self.http_detail_level is not None and \
                self.httplogger.isEnabledFor(logging.DEBUG):
            if self._http_response_headers:
                header_str = \
                    ' '.join('{0}:{1!r}'.format(k, v)
                             for k, v in self._http_response_headers.items())
            else:
                header_str = ''

            if self.http_detail_level == 'summary':
                upayload = ""
            elif self.http_maxlen and (len(payload) > self.http_maxlen):
                upayload = (_ensure_unicode(payload[:self.http_maxlen]) +
                            '...')
            else:
                upayload = _ensure_unicode(payload)

            self.httplogger.debug('Response:%s %s:%s %s %s\n    %s',
                                  self._http_response_conn_id,
                                  self._http_response_status,
                                  self._http_response_reason,
                                  self._http_response_version,
                                  header_str, upayload)

    def record_staged(self):
        """
        Not used for logging. The logs are output in the various
        stage... methods methods
        """
        pass

    def record(self, pywbem_args, pywbem_result, http_request, http_response):
        """Not used for logging"""
        pass


class TestClientRecorder(BaseOperationRecorder):
    """
    An operation recorder that generates test cases for each recorded
    operation.

    **Experimental:** *New in pywbem 0.9 as experimental.*

    The test cases are in the YAML format suitable for the
    `test_client` unit test module of the pywbem project.
    """

    # HTTP header fields to exclude when creating the testcase
    # (in lower case)
    EXCLUDE_REQUEST_HEADERS = [
        'authorization',
        'content-length',
        'content-type',
    ]
    EXCLUDE_RESPONSE_HEADERS = [
        'content-length',
        'content-type',
    ]

    # Dummy server URL and credentials for use in generated test case
    TESTCASE_URL = 'http://acme.com:80'
    TESTCASE_USER = 'username'
    TESTCASE_PASSWORD = 'password'

    def __init__(self, fp):
        """
        Parameters:

          fp (file):
            An open file that each test case will be written to.  This file
            should have been opened in text mode.

            Since there are differences between python 2 and 3 in opening
            files in text mode, the static method
            :meth:`~pywbem.BaseOperationRecorder.open_file`
            can be used to open the file or python 2/3 compatible open::

              from io import open
                  f = open('blah.log', encoding='utf-8')

        Example::

          recorder = TestClientRecorder(
              BaseOperationRecorder.open_file('recorder.log'))
        """
        super(TestClientRecorder, self).__init__()
        self._fp = fp

    def record(self, pywbem_args, pywbem_result, http_request, http_response):
        """
        Function that records the invocation of a single
        :class:`~pywbem.WBEMConnection` method, by appending a corresponding
        test case to the file.

        Parameters: See :meth:`pywbem.BaseOperationRecorder.record`.
        """

        testcase = OrderedDict()
        testcase['name'] = pywbem_args.method
        testcase['description'] = 'Generated by TestClientRecorder'

        tc_pywbem_request = OrderedDict()
        tc_pywbem_request['url'] = TestClientRecorder.TESTCASE_URL
        tc_pywbem_request['creds'] = [TestClientRecorder.TESTCASE_USER,
                                      TestClientRecorder.TESTCASE_PASSWORD]
        tc_pywbem_request['namespace'] = 'root/cimv2'
        tc_pywbem_request['timeout'] = 10
        tc_pywbem_request['debug'] = False
        tc_operation = OrderedDict()
        tc_operation['pywbem_method'] = pywbem_args.method
        for arg_name in pywbem_args.args:
            tc_operation[arg_name] = self.toyaml(pywbem_args.args[arg_name])
        tc_pywbem_request['operation'] = tc_operation
        testcase['pywbem_request'] = tc_pywbem_request

        tc_pywbem_response = OrderedDict()
        if pywbem_result.ret is not None:
            yaml_txt = 'pullresult' if self._pull_op else 'result'
            tc_pywbem_response[yaml_txt] = self.toyaml(pywbem_result.ret)

        if pywbem_result.exc is not None:
            exc = pywbem_result.exc
            if isinstance(exc, CIMError):
                tc_pywbem_response['cim_status'] = self.toyaml(exc.status_code)
            else:
                tc_pywbem_response['exception'] = self.toyaml(
                    exc.__class__.__name__)
        testcase['pywbem_response'] = tc_pywbem_response

        tc_http_request = OrderedDict()
        if http_request is not None:
            tc_http_request['verb'] = http_request.method
            tc_http_request['url'] = TestClientRecorder.TESTCASE_URL
            if http_request.target:
                tc_http_request['url'] += http_request.target
            tc_request_headers = OrderedDict()
            if http_request.headers is not None:
                for hdr_name in http_request.headers:
                    if hdr_name.lower() not in \
                       TestClientRecorder.EXCLUDE_REQUEST_HEADERS:
                        tc_request_headers[hdr_name] = \
                            http_request.headers[hdr_name]
            tc_http_request['headers'] = tc_request_headers
            if http_request.payload is not None:
                data = http_request.payload.decode('utf-8')
                data = data.replace('><', '>\n<').strip()
            else:
                data = None
            tc_http_request['data'] = data
        testcase['http_request'] = tc_http_request

        tc_http_response = OrderedDict()
        if http_response is not None:
            tc_http_response['status'] = http_response.status
            tc_response_headers = OrderedDict()
            if http_response.headers is not None:
                for hdr_name in http_response.headers:
                    if hdr_name.lower() not in \
                       TestClientRecorder.EXCLUDE_RESPONSE_HEADERS:
                        tc_response_headers[hdr_name] = \
                            http_response.headers[hdr_name]
            tc_http_response['headers'] = tc_response_headers
            if http_response.payload is not None:
                data = http_response.payload.decode('utf-8')
                data = data.replace('><', '>\n<').strip()
            else:
                data = None
            tc_http_response['data'] = data
        else:
            tc_http_response['exception'] = "# Change this to a callback " \
                                            "function that causes this " \
                                            "condition."
        testcase['http_response'] = tc_http_response

        testcases = []
        testcases.append(testcase)

        # The file is open in text mode, so we produce a unicode string
        data = yaml.safe_dump(testcases, encoding=None, allow_unicode=True,
                              default_flow_style=False, indent=4)
        data = data.replace('\n\n', '\n')  # YAML dump duplicates newlines
        self._fp.write(data)
        self._fp.flush()

    def toyaml(self, obj):
        """
        Convert any allowable input argument to or return value from an
        operation method to an object that is ready for serialization into
        test_client yaml format.
        """

        # namedtuple is subclass of tuple so it is instance of tuple but
        # not type tuple. Cvt to dictionary and cvt dict to yaml.
        # pylint: disable=unidiomatic-typecheck
        if isinstance(obj, tuple) and type(obj) is not tuple:
            ret_dict = obj._asdict()
            return self.toyaml(ret_dict)
        if isinstance(obj, (list, tuple)):
            ret = []
            # This does not handle namedtuple
            for item in obj:
                ret.append(self.toyaml(item))
            return ret
        elif isinstance(obj, (dict, NocaseDict)):
            ret_dict = OrderedDict()
            for key in obj.keys():  # get keys in original case
                ret_dict[key] = self.toyaml(obj[key])
            return ret_dict
        elif obj is None:
            return obj
        elif isinstance(obj, six.binary_type):
            return obj.decode("utf-8")
        elif isinstance(obj, six.text_type):
            return obj
        elif isinstance(obj, CIMInt):
            # CIMInt is _Longint and therefore may exceed the value range of
            # int in Python 2. Therefore, we convert it to _Longint.
            return _Longint(obj)
        elif isinstance(obj, six.integer_types):
            # This case must be after CIMInt, because CIMInt is _Longint and
            # would match a long value in Python 2.
            # We don't convert six.integer_types to _Longint, because the value
            # fits into the provided type, and there is no need to convert it.
            # Note that in Python 2 (where that would make a difference), int
            # and long do not inherit from each other, so it is likely best if
            # we just don't convert.
            return obj
        elif isinstance(obj, bool):
            return obj
        elif isinstance(obj, CIMFloat):
            return float(obj)
        elif isinstance(obj, CIMDateTime):
            return str(obj)
        elif isinstance(obj, datetime):
            return CIMDateTime(obj)
        elif isinstance(obj, timedelta):
            return CIMDateTime(obj)
        elif isinstance(obj, CIMInstance):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMInstance'
            ret_dict['classname'] = self.toyaml(obj.classname)
            ret_dict['properties'] = self.toyaml(obj.properties)
            ret_dict['path'] = self.toyaml(obj.path)
            return ret_dict
        elif isinstance(obj, CIMInstanceName):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMInstanceName'
            ret_dict['classname'] = self.toyaml(obj.classname)
            ret_dict['namespace'] = self.toyaml(obj.namespace)
            ret_dict['keybindings'] = self.toyaml(obj.keybindings)
            return ret_dict
        elif isinstance(obj, CIMClass):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMClass'
            ret_dict['classname'] = self.toyaml(obj.classname)
            ret_dict['superclass'] = self.toyaml(obj.superclass)
            ret_dict['properties'] = self.toyaml(obj.properties)
            ret_dict['methods'] = self.toyaml(obj.methods)
            ret_dict['qualifiers'] = self.toyaml(obj.qualifiers)
            return ret_dict
        elif isinstance(obj, CIMClassName):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMClassName'
            ret_dict['classname'] = self.toyaml(obj.classname)
            ret_dict['host'] = self.toyaml(obj.host)
            ret_dict['namespace'] = self.toyaml(obj.namespace)
            return ret_dict
        elif isinstance(obj, CIMProperty):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMProperty'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['value'] = self.toyaml(obj.value)
            ret_dict['type'] = self.toyaml(obj.type)
            ret_dict['reference_class'] = self.toyaml(obj.reference_class)
            ret_dict['embedded_object'] = self.toyaml(obj.embedded_object)
            ret_dict['is_array'] = self.toyaml(obj.is_array)
            ret_dict['array_size'] = self.toyaml(obj.array_size)
            ret_dict['class_origin'] = self.toyaml(obj.class_origin)
            ret_dict['propagated'] = self.toyaml(obj.propagated)
            ret_dict['qualifiers'] = self.toyaml(obj.qualifiers)
            return ret_dict
        elif isinstance(obj, CIMMethod):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMMethod'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['return_type'] = self.toyaml(obj.return_type)
            ret_dict['class_origin'] = self.toyaml(obj.class_origin)
            ret_dict['propagated'] = self.toyaml(obj.propagated)
            ret_dict['parameters'] = self.toyaml(obj.parameters)
            ret_dict['qualifiers'] = self.toyaml(obj.qualifiers)
            return ret_dict
        elif isinstance(obj, CIMParameter):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMParameter'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['type'] = self.toyaml(obj.type)
            ret_dict['reference_class'] = self.toyaml(obj.reference_class)
            ret_dict['is_array'] = self.toyaml(obj.is_array)
            ret_dict['array_size'] = self.toyaml(obj.array_size)
            ret_dict['qualifiers'] = self.toyaml(obj.qualifiers)
            return ret_dict
        elif isinstance(obj, CIMQualifier):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMQualifier'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['value'] = self.toyaml(obj.value)
            ret_dict['type'] = self.toyaml(obj.type)
            ret_dict['propagated'] = self.toyaml(obj.propagated)
            ret_dict['tosubclass'] = self.toyaml(obj.tosubclass)
            ret_dict['toinstance'] = self.toyaml(obj.toinstance)
            ret_dict['overridable'] = self.toyaml(obj.overridable)
            ret_dict['translatable'] = self.toyaml(obj.translatable)
            return ret_dict
        elif isinstance(obj, CIMQualifierDeclaration):
            ret_dict = OrderedDict()
            ret_dict['pywbem_object'] = 'CIMQualifierDeclaration'
            ret_dict['name'] = self.toyaml(obj.name)
            ret_dict['type'] = self.toyaml(obj.type)
            ret_dict['value'] = self.toyaml(obj.value)
            ret_dict['is_array'] = self.toyaml(obj.is_array)
            ret_dict['array_size'] = self.toyaml(obj.array_size)
            ret_dict['scopes'] = self.toyaml(obj.scopes)
            ret_dict['tosubclass'] = self.toyaml(obj.tosubclass)
            ret_dict['toinstance'] = self.toyaml(obj.toinstance)
            ret_dict['overridable'] = self.toyaml(obj.overridable)
            ret_dict['translatable'] = self.toyaml(obj.translatable)
            return ret_dict
        else:
            raise TypeError(
                _format("Invalid type in TestClientRecorder.toyaml(): {0} {1}",
                        obj.__class__.__name__, type(obj)))



#
# (C) Copyright 2018 InovaDevelopment.comn
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
# Author: Karl  Schopmeyer <inovadevelopment.com>
#

"""
Mock support for the WBEMConnection class to allow pywbem users to test the
pywbem client without requiring a running WBEM server.

For documentation, see mocksupport.rst.
"""

from __future__ import absolute_import, print_function

from pywbem import MOFWBEMConnection, CIMError, CIM_ERR_INVALID_PARAMETER, \
    CIM_ERR_NOT_FOUND
from pywbem._nocasedict import NocaseDict

from ._resolvermixin import ResolverMixin
from pywbem._utils import _format

# TODO: Use the BaseWBEMConnection and recreate the complete MOFWBEMConnection


class _MockMOFWBEMConnection(MOFWBEMConnection, ResolverMixin):
    """
    Create an adaption of the MOF compiler MOFWBEMConnection class so that to
    directly use the attributes that represent the repository and implement a
    modified CreateClass that resolves the new class. This directs the compiler
    output directly  to the dictionaries for qualifiers, and instances in the
    FakedWBEMConnection object and replaces the CreateClass with a local method
    that allows resolving the created class before it is inserted into the
    repository.

    This class is private to pywbem_mock
    """
    def __init__(self, faked_conn_object):
        """
        Initialize the connection.

          Parameters:

            conn (BaseRepositoryConnection):
              The underlying repository connection.

              `None` means that there is no underlying repository and all
              operations performed through this object will fail.
        """
        super(_MockMOFWBEMConnection, self).__init__(conn=faked_conn_object)

        # Reassign the variables the represent the repository to the
        # faked_conn_object so that we have a common repository
        self.qualifiers = faked_conn_object.qualifiers
        self.instances = faked_conn_object.instances
        self.classes = faked_conn_object.classes

    def CreateClass(self, *args, **kwargs):
        """
        Override the CreateClass method in MOFWBEMConnection

        For a description of the parameters, see
        :meth:`pywbem.WBEMConnection.CreateClass`.
        """
        cc = args[0] if args else kwargs['NewClass']
        namespace = self.getns()

        try:
            self.compile_ordered_classnames.append(cc.classname)

            # The following generates an exception for each new ns
            self.classes[self.default_namespace][cc.classname] = cc
        except KeyError:
            self.classes[namespace] = \
                NocaseDict({cc.classname: cc})

        # Validate that references and embedded instance properties, methods,
        # etc. have classes that exist in repo. This  also institates the
        # mechanism that gets insures that prerequisite classes are inserted
        # into the repo.
        objects = list(cc.properties.values())
        for meth in cc.methods.values():
            objects += list(meth.parameters.values())

        for obj in objects:
            # Validate that reference_class exists in repo
            if obj.type == 'reference':
                try:
                    self.GetClass(obj.reference_class, LocalOnly=True,
                                  IncludeQualifiers=True)
                except CIMError as ce:
                    if ce.status_code == CIM_ERR_NOT_FOUND:
                        raise CIMError(
                            CIM_ERR_INVALID_PARAMETER,
                            _format("Class {0!A} referenced by element {1!A} "
                                    "of class {2!A} in namespace {3!A} does "
                                    "not exist",
                                    obj.reference_class, obj.name,
                                    cc.classname, self.getns()),
                            conn_id=self.conn_id)
                    raise

            elif obj.type == 'string':
                if 'EmbeddedInstance' in obj.qualifiers:
                    eiqualifier = obj.qualifiers['EmbeddedInstance']
                    try:
                        self.GetClass(eiqualifier.value, LocalOnly=True,
                                      IncludeQualifiers=False)
                    except CIMError as ce:
                        if ce.status_code == CIM_ERR_NOT_FOUND:
                            raise CIMError(
                                CIM_ERR_INVALID_PARAMETER,
                                _format("Class {0!A} specified by "
                                        "EmbeddInstance qualifier on element "
                                        "{1!A} of class {2!A} in namespace "
                                        "{3!A} does not exist",
                                        eiqualifier.value, obj.name,
                                        cc.classname, self.getns()),
                                conn_id=self.conn_id)
                        raise

        ccr = self.conn._resolve_class(cc, namespace,
                                       self.qualifiers[namespace])
        if namespace not in self.classes:
            self.classes[namespace] = NocaseDict()
        self.classes[namespace][ccr.classname] = ccr

        try:
            self.class_names[namespace].append(ccr.classname)
        except KeyError:
            self.class_names[namespace] = [ccr.classname]

    def _get_class(self, superclass, namespace=None,
                   local_only=False, include_qualifiers=True,
                   include_classorigin=True):
        """
        This method is just rename of GetClass to support same method
        with both MOFWBEMConnection and FakedWBEMConnection
        """
        return self.GetClass(superclass,
                             namespace=namespace,
                             local_only=local_only,
                             include_qualifiers=include_qualifiers,
                             include_classorigin=include_classorigin)



"""
Unit test for server_definition_file.py
"""

from __future__ import absolute_import, print_function

import os
import six
import pytest
from testfixtures import TempDirectory

from ..utils.pytest_extensions import simplified_test_function
from tests.end2endtest.utils.server_definition_file import ServerDefinition, \
    ServerDefinitionFile, ServerDefinitionFileError


TESTCASES_SERVER_DEFINITION_INIT = [

    # Testcases for ServerDefinition.__init__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * init_args: Tuple of positional arguments to ServerDefinition().
    #   * init_kwargs: Dict of keyword arguments to ServerDefinition().
    #   * exp_attrs: Dict with expected ServerDefinition attributes.
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Basic parameter checking
    (
        "Order of positional parameters",
        dict(
            init_args=(
                'myserver',
                {
                    'description': 'mydesc',
                    'url': 'https://myserver',
                    'user': 'myuser',
                    'password': 'mypassword',
                    'cert_file': 'mycertfile',
                    'key_file': 'mykeyfile',
                    'ca_certs': 'mycacerts',
                    'no_verification': 'myveri',
                },
            ),
            init_kwargs=dict(),
            exp_attrs={
                'nickname': 'myserver',
                'description': 'mydesc',
                'url': 'https://myserver',
                'user': 'myuser',
                'password': 'mypassword',
                'cert_file': 'mycertfile',
                'key_file': 'mykeyfile',
                'ca_certs': 'mycacerts',
                'no_verification': 'myveri',
            },
        ),
        None, None, True
    ),
    (
        "Names of keyword arguments",
        dict(
            init_args=(),
            init_kwargs=dict(
                nickname='myserver',
                server_dict={
                    'description': 'mydesc',
                    'url': 'https://myserver',
                    'user': 'myuser',
                    'password': 'mypassword',
                    'cert_file': 'mycertfile',
                    'key_file': 'mykeyfile',
                    'ca_certs': 'mycacerts',
                    'no_verification': 'myveri',
                },

            ),
            exp_attrs={
                'nickname': 'myserver',
                'description': 'mydesc',
                'url': 'https://myserver',
                'user': 'myuser',
                'password': 'mypassword',
                'cert_file': 'mycertfile',
                'key_file': 'mykeyfile',
                'ca_certs': 'mycacerts',
                'no_verification': 'myveri',
            },
        ),
        None, None, True
    ),

    # Omitted init parameters
    (
        "Omitted optional parameter: description",
        dict(
            init_args=(),
            init_kwargs=dict(
                nickname='myserver',
                server_dict={
                    'url': 'https://myserver',
                    'user': 'myuser',
                    'password': 'mypassword',
                    'cert_file': 'mycertfile',
                    'key_file': 'mykeyfile',
                    'ca_certs': 'mycacerts',
                    'no_verification': 'myveri',
                },

            ),
            exp_attrs={
                'nickname': 'myserver',
                'description': '',
                'url': 'https://myserver',
                'user': 'myuser',
                'password': 'mypassword',
                'cert_file': 'mycertfile',
                'key_file': 'mykeyfile',
                'ca_certs': 'mycacerts',
                'no_verification': 'myveri',
            },
        ),
        None, None, True
    ),
    (
        "Missing required parameter: url",
        dict(
            init_args=(),
            init_kwargs=dict(
                nickname='myserver',
                server_dict={
                    'description': 'mydesc',
                    'user': 'myuser',
                    'password': 'mypassword',
                    'cert_file': 'mycertfile',
                    'key_file': 'mykeyfile',
                    'ca_certs': 'mycacerts',
                    'no_verification': 'myveri',
                },
            ),
            exp_attrs=None,
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Missing required parameter: user",
        dict(
            init_args=(),
            init_kwargs=dict(
                nickname='myserver',
                server_dict={
                    'description': 'mydesc',
                    'url': 'https://myserver',
                    'password': 'mypassword',
                    'cert_file': 'mycertfile',
                    'key_file': 'mykeyfile',
                    'ca_certs': 'mycacerts',
                    'no_verification': 'myveri',
                },
            ),
            exp_attrs=None,
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Missing required parameter: password",
        dict(
            init_args=(),
            init_kwargs=dict(
                nickname='myserver',
                server_dict={
                    'description': 'mydesc',
                    'url': 'https://myserver',
                    'user': 'myuser',
                    'cert_file': 'mycertfile',
                    'key_file': 'mykeyfile',
                    'ca_certs': 'mycacerts',
                    'no_verification': 'myveri',
                },
            ),
            exp_attrs=None,
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Omitted optional parms: cert_file, key_file, ca_certs, no_verif.",
        dict(
            init_args=(),
            init_kwargs=dict(
                nickname='myserver',
                server_dict={
                    'description': 'mydesc',
                    'url': 'https://myserver',
                    'user': 'myuser',
                    'password': 'mypassword',
                },

            ),
            exp_attrs={
                'nickname': 'myserver',
                'description': 'mydesc',
                'url': 'https://myserver',
                'user': 'myuser',
                'password': 'mypassword',
                'cert_file': None,
                'key_file': None,
                'ca_certs': None,
                'no_verification': True,
            },
        ),
        None, None, True
    ),
]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_SERVER_DEFINITION_INIT)
@simplified_test_function
def test_ServerDefinition_init(testcase, init_args, init_kwargs, exp_attrs):
    """
    Test function for ServerDefinition.__init__()
    """

    # The code to be tested
    act_obj = ServerDefinition(*init_args, **init_kwargs)

    # Ensure that exceptions raised in the remainder of this function
    # are not mistaken as expected exceptions
    assert testcase.exp_exc_types is None

    for attr_name in exp_attrs:
        exp_attr_value = exp_attrs[attr_name]
        assert hasattr(act_obj, attr_name), \
            "Missing attribute {0!r} in returned ServerDefinition object". \
            format(attr_name)
        act_attr_value = getattr(act_obj, attr_name)
        assert act_attr_value == exp_attr_value, \
            "Unexpected value for attribute {0!r}: Expected {1!r}, got {2!r}".\
            format(attr_name, exp_attr_value, act_attr_value)


TESTCASES_SERVER_DEFINITION_FILE_INIT = [

    # Testcases for ServerDefinitionFile.__init__()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * sd_file_data: Content of server definition file whose path is passed
    #     to ServerDefinitionFile().
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Basic syntax and semantic checking
    (
        "Invalid YAML syntax: Mixing list and dict",
        dict(
            sd_file_data="servers:\n"
                         " - foo\n"
                         " bar:\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Empty file",
        dict(
            sd_file_data="",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Top level entity is not a dict",
        dict(
            sd_file_data="servers:\n"
                         " - foo\n"
                         " bar:\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "'servers' item with incorrect type list",
        dict(
            sd_file_data="- servers: {}\n"
                         "- server_groups: {}\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Missing 'servers' item",
        dict(
            sd_file_data="server_groups: {}\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "'servers' item with incorrect type string",
        dict(
            sd_file_data="servers: bla\n"
                         "server_groups: {}\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "'server_groups' item with incorrect type list",
        dict(
            sd_file_data="servers: {}\n"
                         "server_groups: []\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "'server_groups' item with incorrect type string",
        dict(
            sd_file_data="servers: {}\n"
                         "server_groups: bla\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Server group grp1 with incorrect type string",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1: srv1\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Server group grp1 with incorrect type dict",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    srv1: {}\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Server group grp1 with one item that has incorrect type dict",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1:\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Server group grp1 with one item that has incorrect type list",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - []\n",
        ),
        ServerDefinitionFileError, None, True
    ),

    # More semantic errors
    (
        "Server group grp1 with one item srv1 but no servers",
        dict(
            sd_file_data="servers: {}\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Server group grp1 with one item srv12 but only server srv1",
        dict(
            sd_file_data="servers: {}\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv12\n",
        ),
        ServerDefinitionFileError, None, True
    ),
    (
        "Server group grp1 with one item srv1 but only server srv12",
        dict(
            sd_file_data="servers: {}\n"
                         "  srv12:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1\n",
        ),
        ServerDefinitionFileError, None, True
    ),

    # Valid but trivial server definition files
    (
        "Valid file with no servers and no server groups",
        dict(
            sd_file_data="servers: {}\n"
                         "server_groups: {}\n",
        ),
        None, None, True
    ),
    (
        "Valid file with no servers omitted 'server_groups' item",
        dict(
            sd_file_data="servers: {}\n",
        ),
        None, None, True
    ),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_SERVER_DEFINITION_FILE_INIT)
@simplified_test_function
def test_ServerDefinitionFile_init(
        testcase, sd_file_data):
    """
    Test function for ServerDefinitionFile.__init__()
    """

    with TempDirectory() as tmp_dir:

        # Create the server definition file
        fd_filename = 'tmp_server_definition_file.yml'
        sd_filepath = os.path.join(tmp_dir.path, fd_filename)
        if isinstance(sd_file_data, six.text_type):
            sd_file_data = sd_file_data.encode('utf-8')
        tmp_dir.write(fd_filename, sd_file_data)

        # The code to be tested
        ServerDefinitionFile(filepath=sd_filepath)


TESTCASES_SERVER_DEFINITION_FILE_LIST_SERVERS = [

    # Testcases for ServerDefinitionFile.list_servers()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * sd_file_data: Content of server definition file whose path is passed
    #     to ServerDefinitionFile().
    #   * list_nickname: Nickname of server or server group for list_servers()
    #   * exp_sd_nick_list: List with the nickname of the expected
    #     ServerDefinition objects returned by
    #     ServerDefinitionFile.list_servers().
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Error cases
    (
        "Listing non-existing nickname with no servers and no server groups",
        dict(
            sd_file_data="servers: {}\n"
                         "server_groups: {}\n",
            list_nickname='foo',
            exp_sd_nick_list=None,
        ),
        ValueError, None, True
    ),

    # Valid cases
    (
        "Listing single server",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv2:\n"
                         "    url: https://srv2\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1\n"
                         "    - srv2\n",
            list_nickname='srv1',
            exp_sd_nick_list=['srv1'],
        ),
        None, None, True
    ),
    (
        "Listing group with two servers",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv2:\n"
                         "    url: https://srv2\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1\n"
                         "    - srv2\n",
            list_nickname='grp1',
            exp_sd_nick_list=['srv1', 'srv2'],
        ),
        None, None, True
    ),
    (
        "Listing nested group (no duplicates)",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv2:\n"
                         "    url: https://srv2\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv3:\n"
                         "    url: https://srv3\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1\n"
                         "    - srv2\n"
                         "  grp2:\n"
                         "    - grp1\n"
                         "    - srv3\n",
            list_nickname='grp2',
            exp_sd_nick_list=['srv1', 'srv2', 'srv3'],
        ),
        None, None, True
    ),
    (
        "Listing nested group with forward ref (no duplicates)",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv2:\n"
                         "    url: https://srv2\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv3:\n"
                         "    url: https://srv3\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp2:\n"
                         "    - grp1\n"
                         "    - srv3\n"
                         "  grp1:\n"
                         "    - srv1\n"
                         "    - srv2\n",
            list_nickname='grp2',
            exp_sd_nick_list=['srv1', 'srv2', 'srv3'],
        ),
        None, None, True
    ),
    (
        "Listing nested group with duplicates in same group",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv2:\n"
                         "    url: https://srv2\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv3:\n"
                         "    url: https://srv3\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1\n"
                         "    - srv2\n"
                         "    - srv1\n"
                         "  grp2:\n"
                         "    - grp1\n"
                         "    - srv3\n",
            list_nickname='grp2',
            exp_sd_nick_list=['srv1', 'srv2', 'srv3'],
        ),
        None, None, True
    ),
    (
        "Listing nested group with duplicates at different levels",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv2:\n"
                         "    url: https://srv2\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv3:\n"
                         "    url: https://srv3\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1\n"
                         "    - srv2\n"
                         "  grp2:\n"
                         "    - grp1\n"
                         "    - srv1\n"
                         "    - srv3\n",
            list_nickname='grp2',
            exp_sd_nick_list=['srv1', 'srv2', 'srv3'],
        ),
        None, None, True
    ),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_SERVER_DEFINITION_FILE_LIST_SERVERS)
@simplified_test_function
def test_ServerDefinitionFile_list_servers(
        testcase, sd_file_data, list_nickname, exp_sd_nick_list):
    """
    Test function for ServerDefinitionFile.list_servers()
    """

    with TempDirectory() as tmp_dir:

        # Create the server definition file
        fd_filename = 'tmp_server_definition_file.yml'
        sd_filepath = os.path.join(tmp_dir.path, fd_filename)
        if isinstance(sd_file_data, six.text_type):
            sd_file_data = sd_file_data.encode('utf-8')
        tmp_dir.write(fd_filename, sd_file_data)

        try:
            sdf = ServerDefinitionFile(filepath=sd_filepath)
        except Exception as exc:
            pytest.fail(
                "Unexpected exception from ServerDefinitionFile(): {0}: {1}".
                format(exc.__class__.__name__, exc))

        # The code to be tested
        sd_list = sdf.list_servers(list_nickname)

        # Ensure that exceptions raised in the remainder of this function
        # are not mistaken as expected exceptions
        assert testcase.exp_exc_types is None

        act_list_servers_len = len(sd_list)
        assert act_list_servers_len == len(exp_sd_nick_list), \
            "Unexpected number of ServerDefinition objects returned from " \
            "list_servers(): Expected nicks {0!r}, got nicks {1!r}". \
            format(exp_sd_nick_list, [sd.nickname for sd in sd_list])

        for i, sd in enumerate(sd_list):
            exp_sd_nick = exp_sd_nick_list[i]
            assert sd.nickname == exp_sd_nick, \
                "Unexpected ServerDefinition object returned from " \
                "list_servers() at position {0}: " \
                "Expected nick {1!r}, got nick {2!r}". \
                format(i, exp_sd_nick, sd.nickname)


TESTCASES_SERVER_DEFINITION_FILE_LIST_ALL_SERVERS = [

    # Testcases for ServerDefinitionFile.list_all_servers()

    # Each list item is a testcase tuple with these items:
    # * desc: Short testcase description.
    # * kwargs: Keyword arguments for the test function:
    #   * sd_file_data: Content of server definition file whose path is passed
    #     to ServerDefinitionFile().
    #   * exp_sd_nick_list: List with the nickname of the expected
    #     ServerDefinition objects returned by
    #     ServerDefinitionFile.list_all_servers().
    # * exp_exc_types: Expected exception type(s), or None.
    # * exp_warn_types: Expected warning type(s), or None.
    # * condition: Boolean condition for testcase to run, or 'pdb' for debugger

    # Error cases
    (
        "Empty servers and server groups",
        dict(
            sd_file_data="servers: {}\n"
                         "server_groups: {}\n",
            exp_sd_nick_list=[],
        ),
        None, None, True
    ),

    # Valid cases
    (
        "Two servers, no groups",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv2:\n"
                         "    url: https://srv2\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1\n"
                         "    - srv2\n",
            exp_sd_nick_list=['srv1', 'srv2'],
        ),
        None, None, True
    ),
    (
        "Two servers, two groups",
        dict(
            sd_file_data="servers:\n"
                         "  srv1:\n"
                         "    url: https://srv1\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "  srv2:\n"
                         "    url: https://srv2\n"
                         "    user: USER\n"
                         "    password: PASSWORD\n"
                         "server_groups:\n"
                         "  grp1:\n"
                         "    - srv1\n"
                         "    - srv2\n"
                         "  grp2:\n"
                         "    - srv2\n",
            exp_sd_nick_list=['srv1', 'srv2'],
        ),
        None, None, True
    ),

]


@pytest.mark.parametrize(
    "desc, kwargs, exp_exc_types, exp_warn_types, condition",
    TESTCASES_SERVER_DEFINITION_FILE_LIST_ALL_SERVERS)
@simplified_test_function
def test_ServerDefinitionFile_list_all_servers(
        testcase, sd_file_data, exp_sd_nick_list):
    """
    Test function for ServerDefinitionFile.list_all_servers()
    """

    with TempDirectory() as tmp_dir:

        # Create the server definition file
        fd_filename = 'tmp_server_definition_file.yml'
        sd_filepath = os.path.join(tmp_dir.path, fd_filename)
        if isinstance(sd_file_data, six.text_type):
            sd_file_data = sd_file_data.encode('utf-8')
        tmp_dir.write(fd_filename, sd_file_data)

        try:
            sdf = ServerDefinitionFile(filepath=sd_filepath)
        except Exception as exc:
            pytest.fail(
                "Unexpected exception from ServerDefinitionFile(): {0}: {1}".
                format(exc.__class__.__name__, exc))

        # The code to be tested
        sd_list = sdf.list_all_servers()

        # Ensure that exceptions raised in the remainder of this function
        # are not mistaken as expected exceptions
        assert testcase.exp_exc_types is None

        act_list_servers_len = len(sd_list)
        assert act_list_servers_len == len(exp_sd_nick_list), \
            "Unexpected number of ServerDefinition objects returned from " \
            "list_all_servers(): Expected nicks {0!r}, got nicks {1!r}". \
            format(exp_sd_nick_list, [sd.nickname for sd in sd_list])

        for i, sd in enumerate(sd_list):
            exp_sd_nick = exp_sd_nick_list[i]
            assert sd.nickname == exp_sd_nick, \
                "Unexpected ServerDefinition object returned from " \
                "list_all_servers() at position {0}: " \
                "Expected nick {1!r}, got nick {2!r}". \
                format(i, exp_sd_nick, sd.nickname)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
    Example of using Iterator Operations to retrieve instances from a
    WBEM Server. This example executes the IterEnumerateInstances and the
    corresponding IterEnumeratePaths operations.

    It:
        Creates a connection
        Opens an enumeration session with OpenEnumerateInstances
        Executes a pull loop until the result.eos =True
        Displays overall statistics on the returns
    It also displays the results of the open and each pull in detail
"""

from __future__ import print_function

import sys
import argparse as _argparse
from pywbem import WBEMConnection, CIMError, Error
from pywbem._cliutils import SmartFormatter as _SmartFormatter

def _str_to_bool(str_):
    """Convert string to bool (in argparse context)."""
    if str_.lower() not in ['true', 'false', 'none']:
        raise ValueError('Need bool; got %r' % str_)
    return {'true': True, 'false': False, 'none': None}[str_.lower()]

def add_nullable_boolean_argument(parser, name, default=None, help_=None):
    """Add a boolean argument to an ArgumentParser instance."""
    group = parser.add_mutually_exclusive_group()
    group.add_argument(
        '--' + name, nargs='?', default=default, const=True, type=_str_to_bool,
        help=help_)
    group.add_argument('--no' + name, dest=name, action='store_false',
                       help=help_)

def create_parser(prog):
    """
    Parse command line arguments, connect to the WBEM server and open the
    interactive shell.
    """

    usage = '%(prog)s [options] server'
    desc = 'Provide an interactive shell for issuing operations against' \
           ' a WBEM server.'
    epilog = """
Examples:
  %s https://localhost:15345 -n vendor -u sheldon -p penny
          - (https localhost, port=15345, namespace=vendor user=sheldon
         password=penny)

  %s http://[2001:db8::1234-eth0] -(http port 5988 ipv6, zone id eth0)
""" % (prog, prog)

    argparser = _argparse.ArgumentParser(
        prog=prog, usage=usage, description=desc, epilog=epilog,
        add_help=False, formatter_class=_SmartFormatter)

    pos_arggroup = argparser.add_argument_group(
        'Positional arguments')
    pos_arggroup.add_argument(
        'server', metavar='server', nargs='?',
        help='R|Host name or url of the WBEM server in this format:\n'
             '    [{scheme}://]{host}[:{port}]\n'
             '- scheme: Defines the protocol to use;\n'
             '    - "https" for HTTPs protocol\n'
             '    - "http" for HTTP protocol.\n'
             '  Default: "https".\n'
             '- host: Defines host name as follows:\n'
             '     - short or fully qualified DNS hostname,\n'
             '     - literal IPV4 address(dotted)\n'
             '     - literal IPV6 address (RFC 3986) with zone\n'
             '       identifier extensions(RFC 6874)\n'
             '       supporting "-" or %%25 for the delimiter.\n'
             '- port: Defines the WBEM server port to be used\n'
             '  Defaults:\n'
             '     - HTTP  - 5988\n'
             '     - HTTPS - 5989\n')

    server_arggroup = argparser.add_argument_group(
        'Server related options',
        'Specify the WBEM server namespace and timeout')
    server_arggroup.add_argument(
        '-n', '--namespace', dest='namespace', metavar='namespace',
        default='root/cimv2',
        help='R|Default namespace in the WBEM server for operation\n'
             'requests when namespace option not supplied with\n'
             'operation request.\n'
             'Default: %(default)s')
    server_arggroup.add_argument(
        '-t', '--timeout', dest='timeout', metavar='timeout', type=int,
        default=None,
        help='R|Timeout of the completion of WBEM Server operation\n'
             'in seconds(integer between 0 and 300).\n'
             'Default: No timeout')

    server_arggroup.add_argument(
        '-c', '--classname', dest='classname', metavar='classname',
        default='CIM_ManagedElement',
        help='R|Classname in the WBEM server for operation\n'
             'request.\n'
             'Default: %(default)s')

    security_arggroup = argparser.add_argument_group(
        'Connection security related options',
        'Specify user name and password or certificates and keys')
    security_arggroup.add_argument(
        '-u', '--user', dest='user', metavar='user',
        help='R|User name for authenticating with the WBEM server.\n'
             'Default: No user name.')
    security_arggroup.add_argument(
        '-p', '--password', dest='password', metavar='password',
        help='R|Password for authenticating with the WBEM server.\n'
             'Default: Will be prompted for, if user name\nspecified.')
    security_arggroup.add_argument(
        '-nvc', '--no-verify-cert', dest='no_verify_cert',
        action='store_true',
        help='Client will not verify certificate returned by the WBEM'
             ' server (see cacerts). This bypasses the client-side'
             ' verification of the server identity, but allows'
             ' encrypted communication with a server for which the'
             ' client does not have certificates.')
    security_arggroup.add_argument(
        '--cacerts', dest='ca_certs', metavar='cacerts',
        help='R|File or directory containing certificates that will be\n'
             'matched against a certificate received from the WBEM\n'
             'server. Set the --no-verify-cert option to bypass\n'
             'client verification of the WBEM server certificate.\n')

    security_arggroup.add_argument(
        '--certfile', dest='cert_file', metavar='certfile',
        help='R|Client certificate file for authenticating with the\n'
             'WBEM server. If option specified the client attempts\n'
             'to execute mutual authentication.\n'
             'Default: Simple authentication.')
    security_arggroup.add_argument(
        '--keyfile', dest='key_file', metavar='keyfile',
        help='R|Client private key file for authenticating with the\n'
             'WBEM server. Not required if private key is part of the\n'
             'certfile option. Not allowed if no certfile option.\n'
             'Default: No client key file. Client private key should\n'
             'then be part  of the certfile')

    general_arggroup = argparser.add_argument_group(
        'General options')
    general_arggroup.add_argument(
        '-v', '--verbose', dest='verbose',
        action='store_true', default=False,
        help='Print more messages while processing')

    general_arggroup.add_argument(
        '-m', '--MaxObjectCount', dest='max_object_count',
        metavar='MaxObjectCount', default=100,
        help='R|Maximum object count in pull responses.\n'
             'Default: 100.')

    add_nullable_boolean_argument(
        general_arggroup, 'usepull',
        help_='Use Pull Operations.  If --usepull, pull will be used. If'
              ' --nousepull will execute Enumerate operation. Default is'
              ' `None` where the system decides.')
    general_arggroup.add_argument(
        '-h', '--help', action='help',
        help='Show this help message and exit')

    return argparser

def parse_cmdline(argparser_):
    """ Parse the command line.  This tests for any required args"""

    opts = argparser_.parse_args()

    if not opts.server:
        argparser_.error('No WBEM server specified')
        return None
    return opts

def main():
    """
        Get arguments and call the execution function
    """
    prog = sys.argv[0]

    arg_parser = create_parser(prog)

    opts = parse_cmdline(arg_parser)

    print('Parameters: server=%s\n username=%s\n namespace=%s\n'
          ' classname=%s\n max_pull_size=%s\n use_pull=%s' %
          (opts.server, opts.user, opts.namespace, opts.classname,
           opts.max_object_count, opts.usepull))

    # call method to connect to the server
    conn = WBEMConnection(opts.server, (opts.user, opts.password),
                          default_namespace=opts.namespace,
                          no_verification=True,
                          use_pull_operations=opts.usepull)

    #Call method to create iterator
    instances = []
    try:
        iter_instances = conn.IterEnumerateInstances(
            opts.classname,
            MaxObjectCount=opts.max_object_count)

        # generate instances with
        for instance in iter_instances:
            print(' %s %s' % (type(instances), type(instance)))
            instances.append(instance)
            print('\npath=%s\n%s' % (instance.path, instance.tomof()))

    # handle exceptions
    except CIMError as ce:
        print('Operation Failed: CIMError: code=%s, Description=%s' %
              (ce.status_code_name, ce.status_description))
        sys.exit(1)
    except Error as err:
        print ("Operation failed: %s" % err)
        sys.exit(1)

    inst_paths = [inst.path for inst in instances]

    paths = []
    try:
        iter_paths = conn.IterEnumerateInstancePaths(
            opts.classname,
            MaxObjectCount=opts.max_object_count)

        # generate instances with
        for path in iter_paths:
            print('path=%s' % path)
            paths.append(path)

        if len(paths) != len(paths):
            print('Error path and enum response sizes different enum=%s path=%s'
                  % (len(paths), len(paths)))
        for path in paths:
            found = False
            for inst_path in inst_paths:                
                if path == inst_path:
                    found = True
                    break
                else:
                    print('No match \n%s\n%s' %(path, inst_path))
            if not found:
                print('Path %s not found in insts' % path)
    # handle exceptions
    except CIMError as ce:
        print('Operation Failed: CIMError: code=%s, Description=%s' %
              (ce.status_code_name, ce.status_description))
        sys.exit(1)
    except Error as err:
        print ("Operation failed: %s" % err)
        sys.exit(1)

    return 0

if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Example of the EnumerateInstances operation. The execution
   arguments may be either internal defaults if less than the full set of
   arguments is supplied on the command line or the supplied command
   line arguments.

   The command line arguments are:
       - server_url: scheme, server name, (optional) port
       - username
       - password
       - namespace
       - classname

   This simple example allows both http and https
   requests but does not allow verification of server cert or
   mutual authentication

   This example demonstrates executing the connection,
   operation request, displaying results and handling exceptions.
"""

from __future__ import print_function
import sys
from pywbem import WBEMConnection, Error, CIMError

# default connection attributes. Used if not all arguments are
# supplied on the command line.
USERNAME = 'blah'
PASSWORD = 'blah'
TEST_CLASS = 'CIM_ComputerSystem'
TEST_NAMESPACE = 'root/cimv2'
SERVER_URL = 'http://localhost'

def execute_request(server_url, creds, namespace, classname):
    """ Open a connection with the server_url and creds, and
        enumerate instances defined by the functions namespace and
        classname arguments.
        Displays either the error return or the mof for instances
        returned.
    """

    print('Requesting url=%s, ns=%s, class=%s' % \
        (server_url, namespace, classname))

    try:
        # Create a connection
        CONN = WBEMConnection(server_url, creds,
                              default_namespace=namespace,
                              no_verification=True)

        #Issue the request to EnumerateInstances on the defined class
        INSTANCES = CONN.EnumerateInstances(classname)

        #Display of characteristics of the result object
        print('instances type=%s len=%s' % (type(INSTANCES),
                                            len(INSTANCES)))
        #display the mof output
        for inst in INSTANCES:
            print('path=%s\n' % inst.path)
            print(inst.tomof())

    # handle any exception
    except Error as err:
        # If CIMError, display CIMError attributes
        if isinstance(err, CIMError):
            print('Operation Failed: CIMError: code=%s, Description=%s' % \
                  (err.status_code_name, err.status_description))
        else:
            print ("Operation failed: %s" % err)
        sys.exit(1)

def main():
    """ Get arguments and call the execution function"""

    if len(sys.argv) < 6:
        print("Usage: %s server_url username password namespace' \
              ' classname" %  sys.argv[0])
        print('Using internal defaults')
        server_url = SERVER_URL
        namespace = TEST_NAMESPACE
        username = USERNAME
        password = PASSWORD
        classname = TEST_CLASS
    else:
        print('Get from input')
        server_url = sys.argv[1]
        namespace = sys.argv[2]
        username = sys.argv[3]
        password = sys.argv[4]
        classname = sys.argv[5]

    # create the credentials tuple for WBEMConnection
    creds = (username, password)

    # call the method to execute the request and display results
    execute_request(server_url, creds, namespace, classname)

    return 0

if __name__ == '__main__':
    sys.exit(main())





"""
    Pywbem wbemcli scriptlet that counts classes and instances in a WBEMServer.

    This class explored the complete server and retrieve all
    namespaces, then all classes, then all instances for each class and
    displays the results.

    It includes an ignore list that ignores selected classes and another that
    will ignore selected namespaces.
"""
from collections import defaultdict
import six

from pywbem import WBEMServer, CIMError, ConnectionError, \
    TimeoutError

# Create the server and save in global SERVER
print('Find total classes and instances for all namespaces:\n')

SERVER = WBEMServer(CONN)

# The following are classes where for some reason the enumerateinstancenanes
# fails in the test version of Pegasus. Typically they cause significant
# failures if queried, i.e kills pegasus. The classes in this list are ignored
# during the scan.
peg_ignore_list = ['PG_User', 'CWS_DirectoryContainsFile_CXX',
                   'PG_Authorization',
                   'TestCMPI_Fail_2', 'TestCMPI_Fail_3', 'TestCMPI_Fail_4',
                   'TST_FaultyInstance', 'CIM_Error',
                   'CIM_InstMethodCall', 'PG_InstMethodIndication',
                   'CIM_Indication', 'CIM_InstIndication', 'PG_EmbeddedError']

# Any namespaces in this list are ignored.
ignore_namespaces_list = []

# list of ignored classes including namespaces
ignored_classes = []
# list of failed classes including namespace
failed_classes = []

# count of total ignored classes, instances, and failures.
total_classes = 0
total_instances = 0
total_errors = 0

table_print_format = '%9s %11s %-30s %5s'

# prepare and print table header with column names
max_ns_len = 20
table_hdr = table_print_format % ('# classes',
                                  '# instances',
                                  'namespace/classname',
                                  '# Errors')
print(table_hdr)
print('=' * len(table_hdr))

# filter namespaces to be ignored
namespaces = [ns
              for ns in sorted(SERVER.namespaces)
              if ns not in ignore_namespaces_list]

# set max namespace length for output table
max_ns_len2 = len(max(namespaces, key=len)) + 5

for ns in namespaces:
    classnames = ecn(ns=ns, di=True)
    ns_error_count = 0
    # this accumulates instance counts across the namespace
    # key is classname, value is total instance count
    inst_dict = defaultdict(int)
    assoc_d = {}

    # build filtered classname list
    classnames = [classname
                  for classname in classnames
                  if classname not in peg_ignore_list]

    for classname in classnames:
        cl = gc(classname, ns=ns)
        assoc_d[classname] = True if 'Association' in cl.qualifiers else False

        try:
            instnames = ein(ns=ns, cn=classname)
            for instname in instnames:
                # filter to count only instances in this class, not subclasses
                if instname.classname == classname:
                    inst_dict[classname] += 1
                    total_instances += len(instnames)
        except CIMError as ce:
            failed_classes.append((classname, ns, ce))
            ns_error_count += 1
        except TimeoutError as te:
            print('Timeout Error ns: %s class: %s err: %s. Quitting' %
                  (ns, classname, te))
            quit(2)
        except ConnectionError as ce:
            print('Connection Error ns: %s class: %s err: %s. Quitting' %
                  (ns, classname, ce))
            quit(2)
        except Exception as exc:
            ns_error_count += 1
            print('Exception error ns %s class %s, err %s. Quitting' %
                  (ns, classname, exc))
            quit(3)

    # display results for this namespace
    print(table_print_format %
          (len(classnames), sum(six.itervalues(inst_dict)), ns, ns_error_count))

    # display the counts per instance. Note that this is NOT the same
    # as ein returns but the instance count for this class alone.
    for classname, inst_count in six.iteritems(inst_dict):
        if inst_count != 0:
            assoc_flag = '(*)' if assoc_d[classname] else "   "
            classname = '%s %s' % (classname, assoc_flag)
            print('%9s %11s %-30s %5s' % ("", inst_count, classname, ""))

        total_classes += len(classnames)
    total_errors += ns_error_count

print('=' * len(table_hdr))
print(table_print_format %
      (len(SERVER.namespaces), total_classes, '', total_errors))

if ignored_classes:
    print('List of classes ignored:')
    for cn in ignored_classes:
        print('  %s' % cn)

if failed_classes:
    print('List of classes That failed with CIMError exception:')
    for item in failed_classes:
        print('  %s:%s, err %s' % (item[0], item[1], item[2].status_code_name))

print('Success. Quit Scriplet. Quiting wbemcli')
quit()
